/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_ANNEAL_H
#define O2SCL_GSL_ANNEAL_H

#include <o2scl/sim_anneal.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_siman.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Multidimensional minimization by simulated annealing (GSL)
      
      Minimize by simulated annealing for an arbitrary temperature
      schedule and random number generator. A move is defined by
      \f[
      x_{i,\mathrm{new}} = \mathrm{stepsize} (2 u_i - 1) + x_{i,\mathrm{old}}
      \f]
      where the \f$u_i\f$ are random numbers between 0 and 1.  The
      random number generator and temperature schedule are set in the
      parent, sim_anneal.  The variables minimize::tolx and
      minimize::tolf are not used

      \todo Implement different stepsizes for the different dimensions

      \future Implement a more general routine which would allow
      the solution of discrete problems like the Traveling Salesman
      problem.
      
  */
  template<class param_t, class func_t, class vec_t=ovector_view,
    class rng_t=gsl_rnga> class gsl_anneal :
    public sim_anneal<param_t,func_t,vec_t,rng_t> {
      public:

      virtual ~gsl_anneal() {}
      
      /** \brief Calculate the minimum \c fmin of \c func w.r.t the 
	  array \c x0 of size \c nvar.
      */
      virtual int mmin(size_t nvar, vec_t &x0, double &fmin, param_t &pa,
		       func_t &func) 
      {
	
	fmin=0.0;

	allocate(nvar);

	double E, new_E, best_E, T;
	int i, iter=0;
	bool done;
	size_t j;

	for(j=0;j<nvar;j++) {
	  x[j]=x0[j];
	  best_x[j]=x0[j];
	}
    
	E=func(nvar,x,pa);
	best_E=E;
  
	T=this->tp->start(0.0,0,x0,NULL);
	done=false;

	while (!done) {
	  for (i=0;i<this->ntrial;++i) {
	    for (j=0;j<nvar;j++) new_x[j]=x[j];
      
	    step(new_x,nvar);
	    new_E=func(nvar,new_x,pa);
      
	    if(new_E<=best_E){
	      for(j=0;j<nvar;j++) best_x[j]=new_x[j];
	      best_E=new_E;
	    }
      
	    // Take the crucial step: see if the new point is accepted
	    // or not, as determined by the boltzman probability
	    if (new_E<E) {
	      for(j=0;j<nvar;j++) x[j]=new_x[j];
	      E=new_E;
	    } else if (this->def_rng.random() < exp(-(new_E-E)/(boltz*T)) ) {
	      for(j=0;j<nvar;j++) x[j]=new_x[j];
	      E=new_E;
	    }
	  }
    
	  if (this->verbose>0) {
	    this->print_iter(nvar,best_x,best_E,iter,T,"gsl_anneal");
	    iter++;
	  }
    
	  // apply the cooling schedule to the temperature
	  T=this->tp->next(0.0,0,x0,NULL);
	  if (this->tp->done(0.0,0,x0,NULL)) {
	    done=true;
	  }
	}
  
	for(j=0;j<nvar;j++) x0[j]=best_x[j];
	fmin=best_E;

	free();
  
	return 0;
      }
      
      /// Return string denoting type ("gsl_anneal")
      virtual const char *type() { return "gsl_anneal"; }

      /// Size of step (default 10.0).
      double step_size;

      /// Boltzmann factor (default 1.0).
      double boltz;

#ifndef DOXYGEN_INTERNAL
      
      protected:
      
      /// \name Storage for present, next, and best vectors
      //@{
      ovector x, new_x, best_x;
      //@}

      /** \brief Allocate memory for a minimizer over \c n dimensions
	  with stepsize \c step and Boltzmann factor \c boltz_factor 
      */
      int allocate(size_t n, double ustep=10.0, double boltz_factor=1.0) {
	x.allocate(n);
	new_x.allocate(n);
	best_x.allocate(n);
	step_size=ustep;
	boltz=boltz_factor;
	return 0;
      }

      /// Free allocated memory
      int free() {
	x.free();
	new_x.free();
	best_x.free();
	return 0;
      }

      /// Make a step to a new attempted minimum
      int step(vec_t &sx, int nvar) {
	for(int i=0;i<nvar;i++) {
	  double u=this->def_rng.random();
	  sx[i]=2.0*u*step_size-step_size+sx[i];
	}
	return 0;
      }
  
#endif

    };

#ifndef DOXYGENP
}
#endif

#endif
