/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SMART_INTERP_H
#define O2SCL_SMART_INTERP_H

#include <o2scl/interp.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Smart interpolation class
      
      This class can semi-intelligently handle arrays which are
      not-well formed outside the interpolation region. In particular,
      if an initial interpolation or derivative calculation fails, the
      arrays are searched for the largest neighborhood around the
      point \c x for which an interpolation or differentiation will
      likely produce a finite result.
      
  */
  template<class vec_t=ovector_view, 
    class rvec_t=ovector_const_reverse, 
    class svec_t=ovector_const_subvector, 
    class srvec_t=ovector_const_subvector_reverse> class smart_interp {
      
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// \name Storage internally created subvectors
    //@{
    const svec_t *sx;
    const svec_t *sy;

    const srvec_t *srx;
    const srvec_t *sry;
    //@}

    /// \name Pointers to interpolation objects
    //@{
    base_interp<vec_t> *rit1;
    base_interp<rvec_t> *rit2;
    base_interp<svec_t> *rit3;
    base_interp<srvec_t> *rit4;
    //@}

#endif

    public:
    
    smart_interp(base_interp<vec_t> &it1, base_interp<rvec_t> &it2,
		 base_interp<svec_t> &it3, base_interp<srvec_t> &it4) {
      sx=0;
      sy=0;
      srx=0;
      sry=0;

      rit1=&it1;
      rit2=&it2;
      rit3=&it3;
      rit4=&it4;
    };

    smart_interp() {
      sx=0;
      sy=0;
      srx=0;
      sry=0;

      rit1=&cit1;
      rit2=&cit2;
      rit3=&cit3;
      rit4=&cit4;
    };
    
    virtual ~smart_interp() {
      delete sx;
      delete sy;
      delete srx;
      delete sry;
    }

    /// \name Default interpolation objects
    //@{
    cspline_interp<vec_t> cit1;
    cspline_interp<rvec_t> cit2;
    cspline_interp<svec_t> cit3;
    cspline_interp<srvec_t> cit4;
    //@}
    
    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double interp(const double x0, size_t n, const vec_t &x, 
			  const vec_t &y) {
      
      bool increasing=true, to_subset=false;
      double ret=0.0;
      int err0, err1, err2;
      
      err_hnd->reset();
      
      // Determine if array is increasing or decreasing

      if (x[0]==x[n-1]) {
	to_subset=true;
      } else if (x[0]>x[n-1]) {
	increasing=false;
      }

      if (increasing) {

	if (n<rit1->min_size) {
	  set_err("Vector size too small in interp().",gsl_edom);
	  return 0.0;
	}
	err0=rit1->alloc(n);
	if (err0==0) {
	  err1=rit1->init(x,y,n);
	  if (err1==0) {
	    err2=rit1->interp(x,y,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit1->free();
	}
	  
      } else {
	
	if (n<rit2->min_size) {
	  set_err("Vector size too small in interp().",gsl_edom);
	  return 0.0;
	}

	rvec_t rx(x);
	rvec_t ry(y);

	err0=rit2->alloc(n);
	if (err0==0) {
	  err1=rit2->init(rx,ry,n);
	  if (err1==0) {
	    err2=rit2->interp(rx,ry,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit2->free();
	}
	  
      }
      
      if (!finite(ret)) to_subset=true;
      
      if (to_subset) {
	
	size_t nn=0;
	find_subset(x0,x0,n,x,y,nn,increasing);
	
	if (nn<=1) {
	  set_err("Interpolation failed in interp().",gsl_efailed);
	  return 0;
	}

	if (increasing) {
	  if (nn>rit3->min_size) {
	    err0=rit3->alloc(nn);
	    if (err0==0) {
	      err1=rit3->init(*sx,*sy,nn);
	      if (err1==0) {
		err2=rit3->interp(*sx,*sy,nn,x0,ret);
	      }		
	      rit3->free();
	    }
	  } else {
	    set_err("Interpolation failed in interp().",gsl_efailed);
	  }
	} else {
	  if (nn>rit4->min_size) {
	    err0=rit4->alloc(nn);
	    if (err0==0) {
	      err1=rit4->init(*srx,*sry,nn);
	      if (err1==0) {
		err2=rit4->interp(*srx,*sry,nn,x0,ret);
	      }
	      rit4->free();
	    }
	  } else {
	    set_err("Interpolation failed in interp().",gsl_efailed);
	  }
	}
	  
      }
    
      return ret;
    }		      
  
    /// Give the value of the derivative \f$ y^{prime}(x=x_0) \f$ .
    virtual double deriv(const double x0, size_t n, const vec_t &x, 
			 const vec_t &y) {
    
      bool increasing=true, to_subset=false;
      double ret=0.0;
      int err0, err1, err2;
      
      err_hnd->reset();
      
      // Determine if array is increasing or decreasing

      if (x[0]==x[n-1]) {
	to_subset=true;
      } else if (x[0]>x[n-1]) {
	increasing=false;
      }

      if (increasing) {

	if (n<rit1->min_size) {
	  set_err("Vector size too small in deriv().",gsl_edom);
	  return 0.0;
	}
	err0=rit1->alloc(n);
	if (err0==0) {
	  err1=rit1->init(x,y,n);
	  if (err1==0) {
	    err2=rit1->deriv(x,y,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit1->free();
	}
	  
      } else {
	
	if (n<rit2->min_size) {
	  set_err("Vector size too small in deriv().",gsl_edom);
	  return 0.0;
	}

	rvec_t rx(x);
	rvec_t ry(y);

	err0=rit2->alloc(n);
	if (err0==0) {
	  err1=rit2->init(rx,ry,n);
	  if (err1==0) {
	    err2=rit2->deriv(rx,ry,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit2->free();
	}
	  
      }
      
      if (!finite(ret)) to_subset=true;
      
      if (to_subset) {
	
	size_t nn=0;
	find_subset(x0,x0,n,x,y,nn,increasing);
	
	if (nn<=1) {
	  set_err("Interpolation failed in deriv().",gsl_efailed);
	  return 0;
	}

	if (increasing) {
	  if (nn>rit3->min_size) {
	    err0=rit3->alloc(nn);
	    if (err0==0) {
	      err1=rit3->init(*sx,*sy,nn);
	      if (err1==0) {
		err2=rit3->deriv(*sx,*sy,nn,x0,ret);
	      }		
	      rit3->free();
	    }
	  } else {
	    set_err("Interpolation failed in deriv().",gsl_efailed);
	  }
	} else {
	  if (nn>rit4->min_size) {
	    err0=rit4->alloc(nn);
	    if (err0==0) {
	      err1=rit4->init(*srx,*sry,nn);
	      if (err1==0) {
		err2=rit4->deriv(*srx,*sry,nn,x0,ret);
	      }
	      rit4->free();
	    }
	  } else {
	    set_err("Interpolation failed in deriv().",gsl_efailed);
	  }
	}
	  
      }
    
      return ret;
    }		      

    /** \brief Give the value of the second derivative  
	\f$ y^{prime \prime}(x=x_0) \f$ .
    */
    virtual double deriv2(const double x0, size_t n, const vec_t &x, 
			  const vec_t &y) {
      
      bool increasing=true, to_subset=false;
      double ret=0.0;
      int err0, err1, err2;
      
      err_hnd->reset();
      
      // Determine if array is increasing or decreasing

      if (x[0]==x[n-1]) {
	to_subset=true;
      } else if (x[0]>x[n-1]) {
	increasing=false;
      }

      if (increasing) {

	if (n<rit1->min_size) {
	  set_err("Vector size too small in deriv2().",gsl_edom);
	  return 0.0;
	}
	err0=rit1->alloc(n);
	if (err0==0) {
	  err1=rit1->init(x,y,n);
	  if (err1==0) {
	    err2=rit1->deriv2(x,y,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit1->free();
	}
	  
      } else {
	
	if (n<rit2->min_size) {
	  set_err("Vector size too small in deriv2().",gsl_edom);
	  return 0.0;
	}

	rvec_t rx(x);
	rvec_t ry(y);

	err0=rit2->alloc(n);
	if (err0==0) {
	  err1=rit2->init(rx,ry,n);
	  if (err1==0) {
	    err2=rit2->deriv2(rx,ry,n,x0,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit2->free();
	}
	  
      }
      
      if (!finite(ret)) to_subset=true;
      
      if (to_subset) {
	
	size_t nn=0;
	find_subset(x0,x0,n,x,y,nn,increasing);
	
	if (nn<=1) {
	  set_err("Interpolation failed in deriv2().",gsl_efailed);
	  return 0;
	}

	if (increasing) {
	  if (nn>rit3->min_size) {
	    err0=rit3->alloc(nn);
	    if (err0==0) {
	      err1=rit3->init(*sx,*sy,nn);
	      if (err1==0) {
		err2=rit3->deriv2(*sx,*sy,nn,x0,ret);
	      }		
	      rit3->free();
	    }
	  } else {
	    set_err("Interpolation failed in deriv2().",gsl_efailed);
	  }
	} else {
	  if (nn>rit4->min_size) {
	    err0=rit4->alloc(nn);
	    if (err0==0) {
	      err1=rit4->init(*srx,*sry,nn);
	      if (err1==0) {
		err2=rit4->deriv2(*srx,*sry,nn,x0,ret);
	      }
	      rit4->free();
	    }
	  } else {
	    set_err("Interpolation failed in deriv2().",gsl_efailed);
	  }
	}
	  
      }
    
      return ret;
    }		      

    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual double integ(const double a, const double b, 
			 size_t n, const vec_t &x, const vec_t &y) {
			   
      bool increasing=true, to_subset=false;
      double ret=0.0;
      int err0, err1, err2;
      
      err_hnd->reset();
      
      // Determine if array is increasing or decreasing

      if (x[0]==x[n-1]) {
	to_subset=true;
      } else if (x[0]>x[n-1]) {
	increasing=false;
      }

      if (increasing) {

	if (n<rit1->min_size) {
	  set_err("Vector size too small in integ().",gsl_edom);
	  return 0.0;
	}
	err0=rit1->alloc(n);
	if (err0==0) {
	  err1=rit1->init(x,y,n);
	  if (err1==0) {
	    err2=rit1->integ(x,y,n,a,b,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit1->free();
	}
	  
      } else {
	
	if (n<rit2->min_size) {
	  set_err("Vector size too small in integ().",gsl_edom);
	  return 0.0;
	}

	rvec_t rx(x);
	rvec_t ry(y);

	err0=rit2->alloc(n);
	if (err0==0) {
	  err1=rit2->init(rx,ry,n);
	  if (err1==0) {
	    err2=rit2->integ(rx,ry,n,a,b,ret);
	    if (err2!=0) to_subset=true;
	  } else {
	    to_subset=true;
	  }
	  rit2->free();
	}
	  
      }
      
      if (!finite(ret)) to_subset=true;
      
      if (to_subset) {
	
	size_t nn=0;
	find_subset(a,b,n,x,y,nn,increasing);
	
	if (nn<=1) {
	  set_err("Interpolation failed in integ().",gsl_efailed);
	  return 0;
	}

	if (increasing) {
	  if (nn>rit3->min_size) {
	    err0=rit3->alloc(nn);
	    if (err0==0) {
	      err1=rit3->init(*sx,*sy,nn);
	      if (err1==0) {
		err2=rit3->integ(*sx,*sy,nn,a,b,ret);
	      }		
	      rit3->free();
	    }
	  } else {
	    set_err("Interpolation failed in integ().",gsl_efailed);
	  }
	} else {
	  if (nn>rit4->min_size) {
	    err0=rit4->alloc(nn);
	    if (err0==0) {
	      err1=rit4->init(*srx,*sry,nn);
	      if (err1==0) {
		err2=rit4->integ(*srx,*sry,nn,a,b,ret);
	      }
	      rit4->free();
	    }
	  } else {
	    set_err("Interpolation failed in integ().",gsl_efailed);
	  }
	}
	  
      }
    
      return ret;
    }		      

#ifndef DOXGYEN_INTERNAL

    protected:

    /// A lookup function for generic vectors
    size_t local_lookup(size_t n, const vec_t &x, double x0) {
      size_t row=0, i=0;
      while(!finite(x[i]) && i<n-1) i++;
      if (i==n-1) {
	return 0;
      }
      double best=x[i], bdiff=fabs(x[i]-x0);
      for(;i<n;i++) {
	if (finite(x[i]) && fabs(x[i]-x0)<bdiff) {
	  row=i;
	  best=x[i];
	  bdiff=fabs(x[i]-x0);
	}
      }
      return row;
    }

    /** \brief Try to find the largest monotonic and finite region
	around the desired location

	\todo After row and row2 are set, check to make sure the 
	entire inside region is monotonic before expanding
    */
    int find_subset(const double a, const double b, size_t sz, const vec_t &x, 
		    const vec_t &y, size_t &nsz, bool &increasing) 
    {
      
      size_t left, right;
      
      // Lookup the point
      size_t row=local_lookup(sz,x,a);
      if (!finite(x[row]) || !finite(y[row])) {
	set_err_ret("find_subset failed to find any valid points.",
		    gsl_notfound);
      }
      size_t row2=local_lookup(sz,x,b);
      if (!finite(x[row2]) || !finite(y[row2])) {
	set_err_ret("find_subset failed to find any valid points.",
		    gsl_notfound);
      }
  
      // Get a neighboring point if necessary
      if (row2==row) {
	row2=row++;
	if (row2>=sz) row2=row-1;
      }

      // Arrange
      if (row<row2) {
	left=row;
	right=row2;
      } else {
	left=row2;
	right=row;
      }
      
      // Establish whether we're increasing or decreasing
      increasing=true;
      if (row2<row && x[row]<x[row2]) increasing=false;
      if (row2>row && x[row2]<x[row]) increasing=false;

      // Need to check region between row and row2 here

      // Increasing case

      if (increasing) {
	
	while(left>0 && x[left-1]<x[left] && 
	      finite(x[left-1]) && finite(y[left-1])) left--;
	while(right<sz-1 && x[right]<x[right+1] &&
	      finite(x[right+1]) && finite(y[right+1])) right++;
	
	nsz=right-left+1;
	
	delete sx;
	delete sy;
	sx=new svec_t(x,left,nsz);
	sy=new svec_t(y,left,nsz);
	
      } else {

	while(left>0 && x[left-1]>x[left] && 
	      finite(x[left-1]) && finite(y[left-1])) left--;
	while(right<sz-1 && x[right]>x[right+1] &&
	      finite(x[right+1]) && finite(y[right+1])) right++;
	
	nsz=right-left+1;
	
	delete srx;
	delete sry;
	srx=new srvec_t(x,left,nsz);
	sry=new srvec_t(y,left,nsz);

      }
      
      return 0;
    }

#endif

  };

  /** 
      \brief Smart interpolation class with pre-specified vectors
      
      This class can semi-intelligently handle arrays which are
      not-well formed outside the interpolation region. In particular,
      if an initial interpolation or derivative calculation fails, the
      arrays are searched for the largest neighborhood around the
      point \c x for which an interpolation or differentiation will
      likely produce a finite result.
      
  */
  template<class vec_t, class svec_t, class alloc_vec_t, class alloc_t>
    class smart_interp_vec {

#ifndef DOXYGEN_INTERNAL
    
  protected:

    /// Storage for internally created subvector
    svec_t *sx;
    /// Storage for internally created subvector
    svec_t *sy;

    /// Pointer to base interpolation object
    base_interp<vec_t> *rit1;
    /// Pointer to base interpolation object
    base_interp<svec_t> *rit2;

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// True if the user-specified x vector is increasing
    bool inc;

    /// Pointer to user-specified vector
    const vec_t *lx;
    /// Pointer to user-specified vector
    const vec_t *ly;

    /// Reversed version of vector
    alloc_vec_t lrx;
    /// Reversed version of vector
    alloc_vec_t lry;

    /// Size of user-specifed vector
    size_t ln;
    
#endif

  public:
      
    /// Create with base interpolation objects \c it and \c rit
    smart_interp_vec(size_t n, const vec_t &x, const vec_t &y) {
      sx=0;
      sy=0;

      rit1=&cit1;
      rit2=&cit2;
      ln=0;
	
      if (n<rit1->min_size) {
	set_err("Vector size too small in o2scl_interp_vec().",gsl_edom);
      } else {

	if (x[0]>x[n-1]) {
	
	  ao.allocate(lrx,n);
	  ao.allocate(lry,n);

	  for(size_t i=0;i<n;i++) {
	    lrx[n-1-i]=x[i];
	    lry[n-1-i]=y[i];
	  }
	  
	  int r1=rit1->alloc(n);
	  if (r1==0) {
	    rit1->init(lrx,lry,n);
	    ln=n;
	    inc=false;
	  }
	    
	} else {
	    
	  int r1=rit1->alloc(n);
	  if (r1==0) {
	    rit1->init(x,y,n);
	    ln=n;
	    lx=&x;
	    ly=&y;
	    inc=true;
	  }
	    
	}

      } 
	

    }

    /// Create with base interpolation objects \c it and \c rit
    smart_interp_vec(base_interp<vec_t> &it1, base_interp<svec_t> &it2,
		     size_t n, const vec_t &x, const vec_t &y) {
      sx=0;
      sy=0;

      rit1=&it1;
      rit2=&it2;
      ln=0;
	
      if (n<rit1->min_size) {
	set_err("Vector size too small in o2scl_interp_vec().",gsl_edom);
      } else {

	if (x[0]>x[n-1]) {
	
	  ao.allocate(lrx,n);
	  ao.allocate(lry,n);

	  for(size_t i=0;i<n;i++) {
	    lrx[n-1-i]=x[i];
	    lry[n-1-i]=y[i];
	  }
	  
	  int r1=rit1->alloc(n);
	  if (r1==0) {
	    rit1->init(lrx,lry,n);
	    ln=n;
	    inc=false;
	  }
	    
	} else {
	    
	  int r1=rit1->alloc(n);
	  if (r1==0) {
	    rit1->init(x,y,n);
	    ln=n;
	    lx=&x;
	    ly=&y;
	    inc=true;
	  }
	    
	}

      } 
	
    }

    virtual ~smart_interp_vec() {
      delete sx;
      delete sy;

      if (ln>0) {
	rit1->free();
	if (inc==false) {
	  ao.free(lrx);
	  ao.free(lry);
	}
      }
    }

    /// Default base interpolation object
    cspline_interp<vec_t> cit1;

    /// Default base interpolation object
    cspline_interp<svec_t> cit2;

    /// Give the value of the function \f$ y(x=x_0) \f$ .
    virtual double interp(const double x0) {
      double ret=0.0;
      int err;
      if (ln>0) {

	if (inc) {

	  err=rit1->interp(*lx,*ly,ln,x0,ret);
	  if (err!=0) {
	    size_t nn=0;
	    find_inc_subset(x0,ln,*lx,*ly,nn);
	    
	    if (nn>1 && nn>=rit2->min_size) {
	      rit2->alloc(nn);
	      rit2->init(*sx,*sy,nn);
	      err=rit2->interp(*sx,*sy,nn,x0,ret);
	      rit2->free();
	    }
	  }

	} else {

	  err=rit1->interp(lrx,lry,ln,x0,ret);
	  if (err!=0) {
	    size_t nn=0;
	    find_inc_subset(x0,ln,lrx,lry,nn);
	    
	    if (nn>1 && nn>=rit2->min_size) {
	      rit2->alloc(nn);
	      rit2->init(*sx,*sy,nn);
	      err=rit2->interp(*sx,*sy,nn,x0,ret);
	      rit2->free();
	    }
	  }

	}

      }

      return ret;
    }
    
    /// Give the value of the derivative \f$ y^{prime}(x=x_0) \f$ .
    virtual double deriv(const double x0) {
      double ret=0.0;
      if (ln>0) {
	if (inc) rit1->deriv(*lx,*ly,ln,x0,ret);
	else rit1->deriv(lrx,lry,ln,x0,ret);
      }
      return ret;
    }		      
    
    /** \brief Give the value of the second derivative  
	\f$ y^{prime \prime}(x=x_0) \f$ .
    */
    virtual double deriv2(const double x0) {
      double ret=0.0;
      if (ln>0) {
	if (inc) rit1->deriv2(*lx,*ly,ln,x0,ret);
	else rit1->deriv2(lrx,lry,ln,x0,ret);
      }
      return ret;
    }		      
    
    /// Give the value of the integral \f$ \int_a^{b}y(x)~dx \f$ .
    virtual double integ(const double x1, const double x2) {
      double ret=0.0;
      if (ln>0) {
	if (inc) rit1->integ(*lx,*ly,ln,x1,x2,ret);
	else rit1->integ(lrx,lry,ln,x1,x2,ret);
      }
      return ret;
    }		      

#ifndef DOXYGEN_INTERNAL

  protected:

    /// A lookup function for generic vectors
    size_t local_lookup(size_t n, const vec_t &x, double x0) {
      size_t row=0, i=0;
      while(!finite(x[i]) && i<n-1) i++;
      if (i==n-1) {
	return 0;
      }
      double best=x[i], bdiff=fabs(x[i]-x0);
      for(;i<n;i++) {
	if (finite(x[i]) && fabs(x[i]-x0)<bdiff) {
	  row=i;
	  best=x[i];
	  bdiff=fabs(x[i]-x0);
	}
      }
      return row;
    }

    /** \brief Try to find the largest monotonic and finite region
	around the desired location
    */
    int find_inc_subset(const double x0, size_t sz, const vec_t &x, 
			const vec_t &y, size_t &nsz) {
      
      size_t row=local_lookup(sz,x,x0), row2=row++;
      size_t left=row, right=row;
      
      if (row2>sz) row2=row-1;

      if (!finite(x[row]) || !finite(y[row])) {
	set_err_ret("find_subset failed to find any valid points.",
		    gsl_notfound);
      }
  
      // Increasing case
      
      while(left>0 && x[left-1]<x[left] && 
	    finite(x[left-1]) && finite(y[left-1])) left--;
      while(right<sz-1 && x[right]<x[right+1] &&
	    finite(x[right+1]) && finite(y[right+1])) right++;
      
      nsz=right-left+1;
      
      delete sx;
      delete sy;
      sx=new svec_t(x,left,nsz);
      sy=new svec_t(y,left,nsz);
      
      return 0;
    }

#endif

  };

  /// sm_interp typedef 
  typedef smart_interp<ovector_view,ovector_const_reverse,
    ovector_const_subvector,ovector_const_subvector_reverse> sm_interp;

  /// sm_interp_vec typedef
  typedef smart_interp_vec<ovector_view,ovector_const_subvector,ovector,
    ovector_alloc> sm_interp_vec;
  
  /** \brief A specialization of smart_interp for C-style double arrays
   */
  template<size_t n> class sma_interp :
  public smart_interp<double[n],array_const_reverse<n>,array_const_subvector,
    array_const_subvector_reverse> {
    
  public:
    
    /// Create with base interpolation objects 
    sma_interp(base_interp<double[n]> &it1, 
	       base_interp<array_const_reverse<n> > &it2,
	       base_interp<array_const_subvector> &it3,
	       base_interp<array_const_subvector_reverse> &it4)
      : smart_interp<double[n],array_const_reverse<n>,
      array_const_subvector,array_const_subvector_reverse>(it1,it2,it3,it4) {
    }
    
      /// Create with default interpolation objects 
      sma_interp() : smart_interp<double[n],array_const_reverse<n>,
	array_const_subvector,array_const_subvector_reverse>() {
      }
    
  };

  /** 
      \brief A specialization of smart_interp_vec for C-style arrays
  */
  template<class arr_t> class sma_interp_vec : 
  public smart_interp_vec<arr_t,array_const_subvector,arr_t,
    array_alloc<arr_t> > {
    
  public:
    
    /// Create with base interpolation object \c it and \c it2
    sma_interp_vec(base_interp<arr_t> &it, 
		   base_interp<array_const_subvector> &it2,
		   size_t n, const arr_t &x, const arr_t &y) :
      smart_interp_vec<arr_t,array_const_subvector,arr_t,
      array_alloc<arr_t> >(it,it2,n,x,y) {
    }

      /// Create with default interpolation objects
      sma_interp_vec(size_t n, const arr_t &x, const arr_t &y) :    
	smart_interp_vec<arr_t,array_const_subvector,arr_t,
	array_alloc<arr_t> >(n,x,y) {
      }
    
  };
  
  
  
#ifndef DOXYGENP
}
#endif

#endif
