/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_COLD_NSTAR_H
#define O2SCL_COLD_NSTAR_H

#include <o2scl/hadronic_eos.h>
#include <o2scl/tov_solve.h>
#include <o2scl/table.h>
#include <o2scl/fermion.h>
#include <o2scl/cern_mroot_root.h>
#include <o2scl/cern_mroot.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/tov_eos.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Naive static cold neutron star
      \nosubgrouping

      This uses hadronic_eos::calc_e() to compute the equation of 
      state of zero-temperature beta-equilibrated neutron star
      matter and tov_solve::mvsr() to compute the mass versus
      radius curve.

      The electron and muon are given masses 
      o2scl_fm_const::mass_electron and o2scl_fm_const::mass_muon,
      respectively.

      The energy density and pressure are both calculated
      in units  \f$ \mathrm{fm}^{-4} \f$  and the baryon
      density in  \f$ \mathrm{fm}^{-3} \f$ 
      
      The condition for Urca is the area of the triangle
      formed by the neutron, proton, and electron Fermi momenta.
      
      Using the definition of the semi-perimeter, 
      \f[
      s \equiv \left( k_{F,n}+k_{F,p}+k_{F,e} \right)/2
      \f]
      Heron's formula gives the triangle area as
      \f[
      a=\sqrt{s(s-k_{F,n})(s-k_{F,p})(s-k_{F,e})}
      \f]
      
      The column in the eos \ref table labeled \c urca
      is  \f$ a^2 \f$ . If this quantity is positive, then
      direct Urca is allowed.
      
      The squared speed of sound (in units of  \f$ c \f$ )
      is calculated by 
      \f[
      c_s^2 = \frac{ d P }{d \varepsilon}
      \f]
      and this is placed in the column labeled \c cs2.

      The adabatic index is calculated by
      \f[
      \Gamma = \frac{ d \ln P} { d \ln \varepsilon}
      \f]
      Note that  \f$ \Gamma \f$  must be greater than  \f$ 4/3 \f$ 
      at the center of the neutron star for stability. (This
      is a necessary, but not sufficient condition.)

      Note that if the speed of sound is non-monotonic, then
      calc_eos() will only record the lowest density for which
      the EOS becomes acausal.
      
      \todo Ensure that the adiabatic index of the central density 
      is greater than 4/3?
      \todo Implement more stability criteria
      \todo Warn if the EOS becomes pure neutron matter?
  */

  class cold_nstar {
  public:

    cold_nstar();

    /// \name Basic operation
    //@{
    /** 
	\brief Set the equation of state
	
        This should be set before calling calc_eos().
    */
    int set_eos(hadronic_eos &he) {
      hep=&he;
      eos_set=true;
      return 0;
    }
    
    /** 
	\brief Calculate the given equation of state
     */
    int calc_eos(double np_0=0.0);

    /** 
	\brief Compute the density at which the direct Urca process is allowed

	This is faster than using calc_eos() since it does nothing
	other than computes the critical density. It does not store
	the equation of state.
     */
    double calc_urca(double np_0=0.0);

    /** \brief Calculate the M vs. R curve
     */
    int calc_nstar();
    //@}

    /** \brief The starting baryon density (default 0.05)
     */
    double nb_start;

    /** \brief The final baryon density (default 2.0)
     */
    double nb_end;

    /** \brief The baryon density stepsize (default 0.01)
     */
    double dnb;

    /** \brief If true, include muons (default false)
     */
    bool include_muons;

    /// \name Output 
    //@{
    /** 
	\brief The smallest baryon density where the pressure decreases

	If this is zero after calling calc_eos(), then 
	the pressure does not decrease in the specified range
	of baryon density
     */
    double min_bad;

    /** 
	\brief The smallest density where Urca becomes allowed

        If this is zero after calling calc_eos(), then direct
	Urca is never allowed.
     */
    double allow_urca;

    /** 
	\brief The smallest density where Urca becomes disallowed
	
        If this is zero after calling calc_eos(), then direct
	Urca is not disallowed at a higher density than 
	it becomes allowed.
     */
    double deny_urca;

    /** 
	\brief The density at which the EOS becomes acausal

        If this is zero, then the EOS is causal at all baryon densities
	in the specified range
     */
    double acausal;

    /** 
	\brief The pressure at which the EOS becomes acausal

        If this is zero, then the EOS is causal at all baryon densities
	in the specified range
     */
    double acausal_pr;

    /** 
	\brief The energy density at which the EOS becomes acausal

        If this is zero, then the EOS is causal at all baryon densities
	in the specified range
     */
    double acausal_ed;

    /** 
	\brief Solver tolerance (default \f$ 10^{-4} \f$)
    */
    double solver_tol;

    /** 
	\brief Get the eos table (after having called calc_eos())
    */
    table &get_eos_results() {
      return eost;
    }
    
    /** 
	\brief Get the results from the TOV (after having called calc_nstar())
    */
    table &get_tov_results() {
      return tp->get_results();
    }

    //@}

    /** 
	\brief Set the neutron and proton 
	
	The default objects are of type \ref fermion, with mass
	o2scl_fm_const::mass_neutron and o2scl_fm_const::mass_proton.
	These defaults will give incorrect results for
	non-relativistic equations of state.
     */
    int set_n_and_p(fermion &n, fermion &p) {
      np=&n;
      pp=&p;
      return 0;
    }

    /// The default neutron
    fermion def_n;

    /// The default proton
    fermion def_p;

    /** 
	\brief Specify the object for solving the TOV equations
	
        The default uses the low-density equation of state with
	tov::verbose=0. In calc_nstar(), the units are set by calling
	tov_solve::set_units().
    */
    int set_tov(tov_solve &ts) {
      tp=&ts;
      return 0;
    }
    
    /** \brief The default TOV equation solver
     */
    tov_solve def_tov;

    /** \brief Set the equation solver for the EOS
     */
    int set_root(root<void *,funct<void *> > &rf) {
      rp=&rf;
      return 0;
    }

    /** 
	\brief The default equation solver for the EOS
    */
    cern_mroot_root<void *,funct<void *> > def_root;

    /// Default EOS object for the TOV solver
    tov_interp_eos def_tov_eos;

#ifndef DOXYGEN_INTERNAL

  protected:

    /// \name The thermodynamic information
    //@{
    thermo hb, h, l;
    //@}

    /// Solve to ensure zero charge in \f$ \beta \f$-equilibrium
    int solve_fun(double x, double &y, void *&vp);

    /// True if equation of state has been set
    bool eos_set;

    /// The electron
    fermion e;

    /// The muon
    fermion mu;

    /// A pointer to the equation of state
    hadronic_eos *hep;

    /// A pointer to the neutron
    fermion *np;

    /// A pointer to the proton
    fermion *pp;

    /// A pointer to the TOV object
    tov_solve *tp;
    
    /// A pointer to the solver
    root<void *,funct<void *> > *rp;

    /// Storage for the EOS table
    table eost;

    /// The baryon density
    double barn;

#endif

  };


#ifndef DOXYGENP
}
#endif

#endif
