/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/had_gibbs_nseos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

had_gibbs_nseos::had_gibbs_nseos() {
  solverp=&def_solver;
  gx.allocate(5);
  px.allocate(1);
  phase_guess=phase1;
  elec.init(o2scl_fm::mass_electron,2.0);
  mu.init(o2scl_fm::mass_muon,2.0);
  def_solver.shrink_step=true;
}

had_gibbs_nseos::~had_gibbs_nseos() {
  gx.free();
  px.free();
}

int had_gibbs_nseos::fixed_chi(double chi, fermion &n1, fermion &p1, 
			       thermo &th1, fermion &n2, fermion &p2, 
			       thermo &th2, double &nb,
			       int type) {
  
  mm_funct_mfptr<had_gibbs_nseos,void *> *mfm;
  if (type==nuclear_matter) {
    mfm=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::fixed_chi_nuc);
  } else if (type==neutron_matter) {
    mfm=new mm_funct_mfptr<had_gibbs_nseos, void *>
      (this,&had_gibbs_nseos::fixed_chi_neut);
  } else {
    mfm=new mm_funct_mfptr<had_gibbs_nseos, void *>
      (this,&had_gibbs_nseos::fixed_chi_nstar);
  }

  chipar cp={&n1,&p1,&n2,&p2,&th1,&th2,chi};
  int ret;

  void *vpx=(void *)&cp;
  if (type==nstar_matter) {
    
    gx[0]=nb;
    gx[1]=n1.n;
    gx[2]=p1.n;
    gx[3]=n2.n;
    gx[4]=p2.n;
    
    ret=solverp->msolve(5,gx,vpx,*mfm);
  
  } else {

    gx[0]=nb;
    gx[1]=n1.n;
    gx[2]=n2.n;
    
    ret=solverp->msolve(3,gx,vpx,*mfm);
  }

  nb=gx[0];
  
  if (ret!=0) add_err_ret("Solve failed in had_gibbs_nseos::fixed_chi().",
			  ret);
  return 0;
}

int had_gibbs_nseos::calc_e(double nb, fermion &n1, fermion &p1, thermo &th1,
			    fermion &n2, fermion &p2, thermo &th2, 
			    double &chi, int type) {
  
  int ret;
      
  solvepar sp={&n1,&p1,&n2,&p2,&th1,&th2,nb};
  void *vpx=(void *)&sp;

  mm_funct_mfptr<had_gibbs_nseos,void *> *mf1=0, *mf2=0, *mfm;
  if (type==nuclear_matter) {
    mfm=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::gibbs_nuc);
  } else if (type==neutron_matter) {
    mfm=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::gibbs_neut);
  } else {
    mf1=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::phase1_nstar);
    mf2=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::phase2_nstar);
    mfm=new mm_funct_mfptr<had_gibbs_nseos,void *>
      (this,&had_gibbs_nseos::gibbs_nstar);
  }

  // Begin by assuming that we're in phase 1
  if (phase_guess==phase1) {
	
    if (type==nstar_matter) {

      px[0]=n1.n;
      ret=solverp->msolve(1,px,vpx,*mf1);
      if (ret!=0) {
	add_err("Phase 1 solution failed in calc_e(). Continuing.",ret);
      }

    } else {

      if (type==nuclear_matter) n1.n=nb/2.0;
      else n1.n=nb;
      p1.n=nb-n1.n;

      hep1->calc_e(n1,p1,th1);
    }
	
    n2.mu=n1.mu;
    p2.mu=p1.mu;
	
    ret=hep2->calc_p(n2,p2,th2);
    if (ret!=0) {
      add_err("Phase 2 calc_p() failed in calc_e(). Continuing.",ret);
    }

    if (ret==0 && th2.pr>th1.pr) {

      // The pressure can be increased by adding phase2
      
      if (type==nstar_matter) {

	gx[0]=0.01;
	gx[1]=n1.n;
	gx[2]=p1.n;
	gx[3]=n2.n;
	gx[4]=p2.n;
	
	ret=solverp->msolve(5,gx,vpx,*mfm);

      } else {

	gx[0]=0.01;
	gx[1]=n1.n;
	gx[2]=n2.n;
	
	ret=solverp->msolve(3,gx,vpx,*mfm);

      }

      chi=gx[0];
      
      if (ret!=0 && gx[0]>0.9) {
	// Chi is large, so it might be pure phase 2

	if (type==nstar_matter) {

	  px[0]=n2.n;
	  ret=solverp->msolve(1,px,vpx,*mf2);
	  
	  if (ret!=0) {
	    add_err_ret("Phase 2 solution failed in calc_e().",ret);
	  }

	} else {

	  if (type==nuclear_matter) n2.n=nb/2.0;
	  else n2.n=nb;
	  p2.n=nb-n2.n;
	  hep2->calc_e(n2,p2,th2);
	  
	}

	phase_guess=phase2;
	chi=1.0;

      } else if (ret!=0) {
	    
	phase_guess=mixed_phase;
	add_err_ret("Mixed phase solution failed in calc_e().",ret);
	    
      } else {

	phase_guess=mixed_phase;
	
      }

    } else {

      phase_guess=phase1;
      chi=0.0;

    }

  } else if (phase_guess==phase2) {

    if (type==nstar_matter) {
      
      px[0]=n2.n;
      ret=solverp->msolve(1,px,vpx,*mf2);
      if (ret!=0) {
	add_err("Phase 2 solution failed in calc_e(). Continuing.",ret);
      }
      
    } else {

      if (type==nuclear_matter) n2.n=nb/2.0;
      else n2.n=nb;
      p2.n=nb-n2.n;
      hep2->calc_e(n2,p2,th2);

    }
	
    n1.mu=n2.mu;
    p1.mu=p2.mu;
    
    ret=hep1->calc_p(n1,p1,th1);
    if (ret!=0) {
      add_err("Phase 1 calc_p() failed in calc_e(). Continuing.",ret);
    }
	
    if (th1.pr>th2.pr) {
      // The energy can be decreased by adding phase1
	
      if (type==nstar_matter) {

	gx[0]=0.99;
	gx[1]=n1.n;
	gx[2]=p1.n;
	gx[3]=n2.n;
	gx[4]=p2.n;
	
	ret=solverp->msolve(5,gx,vpx,*mfm);

      } else {

	gx[0]=0.99;
	gx[1]=n1.n;
	gx[2]=n2.n;

	ret=solverp->msolve(3,gx,vpx,*mfm);

      }

      chi=gx[0];

      if (ret!=0 && gx[0]<0.1) {
	// Chi is small, so it might be pure phase 1
	    
	px[0]=n1.n;
	ret=solverp->msolve(1,px,vpx,*mf1);

	phase_guess=phase1;
	chi=0.0;
	if (ret!=0) {
	  add_err_ret("Phase 1 solution failed in calc_e().",ret);
	}
	    
      } else {

	phase_guess=mixed_phase;
	if (ret!=0) {
	  add_err_ret("Mixed phase solution failed in calc_e().",ret);
	}

      }

    } else {

      phase_guess=phase2;
      chi=1.0;

    }

  } else {

    if (type==nstar_matter) {
      
      gx[0]=chi;
      gx[1]=n1.n;
      gx[2]=p1.n;
      gx[3]=n2.n;
      gx[4]=p2.n;
      
      ret=solverp->msolve(5,gx,vpx,*mfm);

    } else {
      
      gx[0]=chi;
      gx[1]=n1.n;
      gx[2]=n2.n;
      
      ret=solverp->msolve(3,gx,vpx,*mfm);
      
    }

    chi=gx[0];
    
    if (ret!=0 && gx[0]<0.1) {

      if (type==nstar_matter) {

	px[0]=n1.n;
	ret=solverp->msolve(1,px,vpx,*mf1);
	
	if (ret!=0) {
	  add_err_ret("Phase 1 solution failed in calc_e().",ret);
	} else {
	  phase_guess=phase1;
	  chi=0.0;
	}

      } else {

	if (type==nuclear_matter) n1.n=nb/2.0;
	else n1.n=nb;
	p1.n=nb-n1.n;
	hep1->calc_e(n1,p1,th1);
	phase_guess=phase1;
	chi=0.0;
	
      }
	
    } else if (ret!=0 && gx[0]>0.9) {
      

      if (type==nstar_matter) {
	px[0]=n2.n;
	ret=solverp->msolve(1,px,vpx,*mf2);
	
	if (ret!=0) {
	  add_err_ret("Phase 2 solution failed in calc_e().",ret);
	} else {
	  phase_guess=phase2;
	  chi=1.0;
	}
      } else {

	if (type==nuclear_matter) n2.n=nb/2.0;
	else n2.n=nb;
	p2.n=nb-n2.n;
	hep2->calc_e(n2,p2,th2);
	phase_guess=phase2;
	chi=0.0;
	
      }
	  
    } else {

      phase_guess=mixed_phase;
      if (ret!=0) {
	add_err_ret("Mixed phase solution failed in calc_e().",ret);
      }

    }
	
  }

  return 0;
}

int had_gibbs_nseos::phase1_nstar(size_t nv, const ovector_view &x, 
				  ovector_view &y, 
				  void *&pa) {
  
  solvepar *sp=(solvepar *)pa;
  
  sp->n1->n=x[0];
  if (x[0]<0.0) set_err_ret("Density less than zero in phase1_nstar().",1);
  sp->p1->n=sp->nb-x[0];
  int ret=hep1->calc_e(*sp->n1,*sp->p1,*sp->th1);
  if (ret!=0) {
    set_err_ret
      ("Function calc_e() failed for Phase 1 in phase1_nstar().",ret);
  }
  elec.mu=sp->n1->mu-sp->p1->mu;
  elec.calc_mu_zerot();
      
  y[0]=sp->p1->n-elec.n;

  return 0;
}
     
int had_gibbs_nseos::phase2_nstar(size_t nv, const ovector_view &x, 
				  ovector_view &y, 
				  void *&pa) {
  int ret;
  
  solvepar *sp=(solvepar *)pa;
  
  sp->n2->n=x[0];
  if (x[0]<0.0) set_err_ret("Density less than zero in phase2_nstar().",1);
  sp->p2->n=sp->nb-x[0];
  ret=hep2->calc_e(*sp->n2,*sp->p2,*sp->th2);
  if (ret!=0) {
    set_err_ret
      ("Function calc_e() failed for Phase 2 in phase1_nstar().",ret);
  }
  elec.mu=sp->n2->mu-sp->p2->mu;
  elec.calc_mu_zerot();
      
  y[0]=sp->p2->n-elec.n;

  return 0;
}
     
int had_gibbs_nseos::gibbs_nstar(size_t nv, const ovector_view &x, 
				 ovector_view &y, 
				 void *&pa) {
  int ret;

  solvepar *sp=(solvepar *)pa;

  double chi=x[0];

  if (chi>1.0 || chi<0.0) {
    set_err_ret("Chi out of range in gibbs_nstar().",1);
  }
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in gibbs_nstar().",2);
  }

  sp->n1->n=x[1];
  sp->p1->n=x[2];
  ret=hep1->calc_e(*sp->n1,*sp->p1,*sp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 1 in gibbs_nstar().",ret);

  sp->n2->n=x[3];
  sp->p2->n=x[4];
  ret=hep2->calc_e(*sp->n2,*sp->p2,*sp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in gibbs_nstar().",ret);

  elec.mu=sp->n1->mu-sp->p1->mu;
  elec.calc_mu_zerot();
      
  y[0]=(1.0-chi)*(sp->n1->n+sp->p1->n)+chi*(sp->n2->n+sp->p2->n)-sp->nb;
  y[1]=sp->th1->pr-sp->th2->pr;
  y[2]=(1.0-chi)*sp->p1->n+chi*sp->p2->n-elec.n;
  y[3]=sp->n2->mu-sp->n1->mu;
  y[4]=sp->p2->mu-sp->p1->mu;
      
  return 0;
}

int had_gibbs_nseos::gibbs_nuc(size_t nv, const ovector_view &x, 
			       ovector_view &y, 
			       void *&pa) {
  int ret;

  solvepar *sp=(solvepar *)pa;

  double chi=x[0];

  if (chi>1.0 || chi<0.0) {
    set_err_ret("Chi out of range in gibbs_nuc().",1);
  }
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in gibbs_nuc().",2);
  }

  sp->n1->n=x[1];
  sp->p1->n=x[1];
  ret=hep1->calc_e(*sp->n1,*sp->p1,*sp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 1 in gibbs_nuc().",ret);

  sp->n2->n=x[2];
  sp->p2->n=x[2];
  ret=hep2->calc_e(*sp->n2,*sp->p2,*sp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in gibbs_nuc().",ret);

  elec.mu=sp->n1->mu-sp->p1->mu;
  elec.calc_mu_zerot();
      
  y[0]=(1.0-chi)*(sp->n1->n+sp->p1->n)+chi*(sp->n2->n+sp->p2->n)-sp->nb;
  y[1]=sp->th1->pr-sp->th2->pr;
  y[2]=sp->n2->mu-sp->n1->mu;
      
  return 0;
}

int had_gibbs_nseos::gibbs_neut(size_t nv, const ovector_view &x, 
				ovector_view &y, 
				void *&pa) {
  int ret;

  solvepar *sp=(solvepar *)pa;

  double chi=x[0];

  if (chi>1.0 || chi<0.0) {
    set_err_ret("Chi out of range in gibbs_neut().",1);
  }
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in gibbs_neut().",2);
  }

  sp->n1->n=x[1];
  sp->p1->n=0.0;
  ret=hep1->calc_e(*sp->n1,*sp->p1,*sp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 1 in gibbs_neut().",ret);

  sp->n2->n=x[2];
  sp->p2->n=0.0;
  ret=hep2->calc_e(*sp->n2,*sp->p2,*sp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in gibbs_neut().",ret);

  elec.mu=sp->n1->mu-sp->p1->mu;
  elec.calc_mu_zerot();
      
  y[0]=(1.0-chi)*(sp->n1->n+sp->p1->n)+chi*(sp->n2->n+sp->p2->n)-sp->nb;
  y[1]=sp->th1->pr-sp->th2->pr;
  y[2]=sp->n2->mu-sp->n1->mu;
      
  return 0;
}

int had_gibbs_nseos::fixed_chi_nstar(size_t nv, const ovector_view &x, 
				     ovector_view &y, 
				     void *&pa) {
  int ret;

  chipar *cp=(chipar *)pa;
  
  double nb=x[0];
  
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in fixed_chi_nstar().",2);
  }

  cp->n1->n=x[1];
  cp->p1->n=x[2];
  ret=hep1->calc_e(*cp->n1,*cp->p1,*cp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_nstar().",
		 ret);

  cp->n2->n=x[3];
  cp->p2->n=x[4];
  ret=hep2->calc_e(*cp->n2,*cp->p2,*cp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_nstar().",
		 ret);

  elec.mu=cp->n1->mu-cp->p1->mu;
  elec.calc_mu_zerot();
  
  y[0]=(1.0-cp->chi)*(cp->n1->n+cp->p1->n)+cp->chi*(cp->n2->n+cp->p2->n)-nb;
  y[1]=cp->th1->pr-cp->th2->pr;
  y[2]=(1.0-cp->chi)*cp->p1->n+cp->chi*cp->p2->n-elec.n;
  y[3]=cp->n2->mu-cp->n1->mu;
  y[4]=cp->p2->mu-cp->p1->mu;
  
  return 0;
}

int had_gibbs_nseos::fixed_chi_nuc(size_t nv, const ovector_view &x, 
				   ovector_view &y, 
				   void *&pa) {
  int ret;

  chipar *cp=(chipar *)pa;
  
  double nb=x[0];
  
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in fixed_chi_nuc().",2);
  }

  cp->n1->n=x[1];
  cp->p1->n=x[1];
  ret=hep1->calc_e(*cp->n1,*cp->p1,*cp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_nuc().",
		 ret);

  cp->n2->n=x[2];
  cp->p2->n=x[2];
  ret=hep2->calc_e(*cp->n2,*cp->p2,*cp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_nuc().",
		 ret);

  elec.mu=cp->n1->mu-cp->p1->mu;
  elec.calc_mu_zerot();
  
  y[0]=(1.0-cp->chi)*(cp->n1->n+cp->p1->n)+cp->chi*(cp->n2->n+cp->p2->n)-nb;
  y[1]=cp->th1->pr-cp->th2->pr;
  y[2]=cp->n2->mu-cp->n1->mu;
  // Assume that if the neutron chemical potential, then the proton
  // chemical potentials are also the same
  //  y[3]=cp->p2->mu-cp->p1->mu;
  
  return 0;
}

int had_gibbs_nseos::fixed_chi_neut(size_t nv, const ovector_view &x, 
				    ovector_view &y, 
				    void *&pa) {
  int ret;

  chipar *cp=(chipar *)pa;
  
  double nb=x[0];
  
  if (x[1]<0.0 || x[2]<0.0 || x[3]<0.0 || x[4]<0.0) {
    set_err_ret("Density less than zero in fixed_chi_neut().",2);
  }

  cp->n1->n=x[1];
  cp->p1->n=0.0;
  ret=hep1->calc_e(*cp->n1,*cp->p1,*cp->th1);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_neut().",
		 ret);

  cp->n2->n=x[2];
  cp->p2->n=0.0;
  ret=hep2->calc_e(*cp->n2,*cp->p2,*cp->th2);
  if (ret!=0) set_err_ret
		("Function calc_e() failed for Phase 2 in fixed_chi_neut().",
		 ret);

  elec.mu=cp->n1->mu-cp->p1->mu;
  elec.calc_mu_zerot();
  
  y[0]=(1.0-cp->chi)*(cp->n1->n+cp->p1->n)+cp->chi*(cp->n2->n+cp->p2->n)-nb;
  y[1]=cp->th1->pr-cp->th2->pr;
  y[2]=cp->n2->mu-cp->n1->mu;
  
  return 0;
}

