/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_SKYRME_EOS_H
#define O2SCL_SKYRME_EOS_H

#include <iostream>
#include <string>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/mroot.h>
#include <o2scl/hadronic_eos.h>
#include <o2scl/part.h>
#include <o2scl/collection.h>
#include <o2scl/nonrel_fermion.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Skyrme hadronic equation of state at zero temperature

      Quantities which have units containing powers of energy
      are divided by \f$\hbar c\f$ to ensure all quantities are
      in units of \f$fm\f$. The \f$x_i\f$ and \f$\alpha\f$ are
      unitless, while the original units of the \f$t_i\f$ are:
      - \f$t_0\f$ - \f$\mathrm{MeV}\f$ \f$\mathrm{fm}^3\f$
      - \f$t_1\f$ - \f$\mathrm{MeV}\f$ \f$\mathrm{fm}^5\f$
      - \f$t_2\f$ - \f$\mathrm{MeV}\f$ \f$\mathrm{fm}^5\f$
      - \f$t_3\f$ - \f$\mathrm{MeV}\f$ \f$\mathrm{fm}^{3(1+\alpha)}\f$
      
      These are stored internally with units of:
      - \f$t_0\f$ - \f$\mathrm{fm}^2\f$
      - \f$t_1\f$ - \f$\mathrm{fm}^4\f$
      - \f$t_2\f$ - \f$\mathrm{fm}^4\f$
      - \f$t_3\f$ - \f$\mathrm{fm}^{2+3 \alpha}\f$
      
      The class skyrme_eos_io uses o2scl_const::hc_mev_fm for I/O so that
      all files contain the parameters in the original units.
      
      The chemical potentials do include the rest mass energy.
      
      n.ed and p.ed contain tau_n and tau_p which are in fm^{-5}, not
      fm^{-4}.
      
      The functions for the usual saturation properties are based 
      partly on \ref Brack85.
      
      The Hamiltonian is defined below. \f$a=0,b=1\f$ gives the standard
      definition of the Skyrme Hamiltonian (ref?), while \f$a=1,b=0\f$
      contains the modifications suggested by \ref Osni94.
      Check also \ref Dutta86.
      
      \f[
      {\cal H} = 
      {\cal H}_{k1} +
      {\cal H}_{k2} +
      {\cal H}_{k3} +
      {\cal H}_{p1} +
      {\cal H}_{p2} +
      {\cal H}_{p3} +
      {\cal H}_{g1} +
      {\cal H}_{g2}
      \f]
      
      The kinetic terms are:
      
      \f[
      {\cal H}_{k1} = \frac{\tau_n}{2 m_n} +
      \frac{\tau_p}{2 m_p} 
      \f]
      
      \f[
      {\cal H}_{k2} =
      n \left(\tau_n + \tau_p \right) \left[ \frac{t_1}{4} 
      \left( 1 + \frac{x_1}{2} \right)
      + \frac{t_2}{4} \left( 1 + \frac{x_2}{2} \right) \right]
      \f]
      
      \f[
      {\cal H}_{k3} =
      \left( \tau_n n_n + \tau_p n_p \right) \left[ \frac{t_2}{4} 
      \left( \frac{1}{2} + x_2 \right)
      - \frac{t_1}{4} \left( \frac{1}{2} + x_1 \right) \right]
      \f]
      
      The potential terms are:
      
      \f[
      {\cal H}_{p1} = 
      \frac{t_0}{2} 
      \left[ \left( 1 + \frac{x_0}{2} \right) n^2 - 
      \left( {\textstyle \frac{1}{2}} + x_0 \right) 
      \left( n_n^2 + n_p^2 \right) \right] 
      \f]
      
      \f[
      {\cal H}_{p2} = 
      \frac{a t_3}{6} \left[ \left( 1 + \frac{x_3}{2} \right) n^{\alpha} 
      n_n n_p + 2^{\alpha-2} \left(1 - x_3\right)
      \left(n_n^{\alpha+2} + n_p^{\alpha+2}\right) \right] 
      \f]
      
      \f[
      {\cal H}_{p3} = 
      \frac{b t_3}{12} \left[ \left(1 + \frac{x_3}{2} \right) n^{\alpha+2} -
      \left(\frac{1}{2} + x_3 \right) n^{\alpha} 
      \left( n_n^2+n_p^2 \right) \right]
      \f]
      
      The gradient terms are displayed here for completeness even though
      they are not computed in the code:
      
      \f[
      {\cal H}_{g1} = 
      \frac{3}{32} \left[ t_1 \left(1 - x_1 \right) - 
      t_2 \left(1 + x_2 \right) \right] \left[ \left( \nabla n_n\right)^2 + 
      \left( \nabla n_p \right)^2 \right] 
      \f]
      
      \f[
      {\cal H}_{g2} = 
      \frac{1}{8} \left[ 3 t_1 \left( 1 + 
      \frac{x_1}{2} \right) - t_2 \left(1 + \frac{x_2}{2} \right) \right] 
      \nabla n_n \nabla n_p
      \f]

      A couple useful definitions:
      
      \f[
      t_3^{\prime} = \left(a + b\right) t_3
      \f]
      
      \f[
      C = \frac{3 }{10 m} \left( \frac{3 \pi^2 }{2} \right)^{2/3} 
      \f]
      
      \f[
      \beta = \frac{M}{2} \left[ \frac{1}{4} \left( 3 t_1 + 5 t_2 \right)
      + t_2 x_2 \right] \\
      \f]
      
      Models are taken from the references: \ref Friedrich86, 
      \ref Dutta86, \ref VanGiai81, \ref Chabanat95, \ref Chabanat97, 
      \ref Beiner75, \ref Reinhard95, \ref Bartel79, \ref Dobaczewski94, 
      \ref Osni94, \ref Tondeur84 and others.
      
      \htmlonly
      See Mathematica notebook at 
      <a href="skyrme_eos.nb">
      skyrme_eos.nb</a>, and
      <a href="skyrme_eos.ps">
      skyrme_eos.ps</a>.
      \endhtmlonly
      \latexonly
      See Mathematica notebook at 
      \begin{verbatim}
      doc/o2scl/extras/skyrme_eos.nb
      doc/o2scl/extras/skyrme_eos.ps
      \end{verbatim}
      \endlatexonly

      <b>Finite temperature documentation</b>

      The finite temperature extension is performed using the method
      \ref Prakash97 (see also the Windsurfing the Fermi Sea
      proceedings by Prakash).

      Returned chemical potentials include the rest mass, and energy 
      densities include the rest mass energy density.

      The variables n.nu and p.nu contain the expressions (-mu_n+V_n)/temper
      and (-mu_p+V_p)/temper respectively, where V is the potential part
      of the single particle energy for particle i (i.e. the derivative
      of the Hamiltonian wrt density while energy density held constant).
      Equivalently, n.nu is just -n.kf^2/2/mstar.
     
      For y>4 we use a non-degenerate expansion, and for y<-40 we
      use a Sommerfeld expansion. For y>50 the code fails, but this is well
      outside the regime of applicability of this model.
     
      Runs the zero temperature code if temper<=0.0.

      The finite temperature code does not include attempt to include
      antiparticles and uses part::calc_density().
      
      \note Since this EOS uses the effective masses and chemical
      potentials in the fermion class, the values of
      part::non_interacting for neutrons and protons are set to false
      in many of the functions.

      <b>Todos</b>

      \todo
      - Make sure that this class properly handles particles for which 
      inc_rest_mass is true/false
      - What about the spin-orbit units?
      - Need to write a function that calculates saturation density?
      - Remove use of mnuc in calparfun()?
      - The compressibility could probably use some simplification
      - Make sure the finite-temperature part is properly tested
      
  */
  class skyrme_eos : public hadronic_eos {
  public:

    double t0,t1,t2,t3,x0,x1,x2,x3,alpha,a,b;

    /** 
	\brief Spin-orbit splitting

	This is unused, but included for possible future use and
	present in the internally stored models.
    */
    double W0;

    skyrme_eos();

    /** 
	\brief Load the model named 'model'

	See comments under skyrme_eos::load().
    */
    skyrme_eos(std::string model);

    virtual ~skyrme_eos() {};

    /// Equation of state as a function of densities
    virtual int calc_temp_e(fermion &ne, fermion &pr, const double temper, 
			    thermo &th);

    /// Equation of state as a function of density.
    virtual int calc_e(fermion &ne, fermion &pr, thermo &lt);

    /** 
	\name Saturation properties

	These calculate the various saturation properties exactly from
	the parameters at any density. These routines often assume that 
	the neutron and proton masses are equal.
    */
    //@{
  
    /** 
	\brief Calculate binding energy
      
	\f[
	\frac{E}{A} = C n_B^{2/3} \left( 1 + \beta n_B \right) + 
	\frac{3 t_0}{8} n_B + \frac{t_3^{\prime}}{16} n_B^{\alpha+1} 
	\f]
    */
    virtual double feoa(double nb);
  
    /** 
	\brief Calculate effective mass
      
	\f[
	M^{*}/M = \left(1+ \beta n_B \right)^{-1} \\
	\f]
    */
    virtual double fmsom(double nb);

    /** 
	\brief Calculate compressibility

	\f[
	K = 10 C n_B^{2/3} + \frac{27}{4} t_0 n_B + 40 C \beta n_B^{5/3} + 
	\frac{9 t_3^{\prime}}{16} 
	\alpha \left( \alpha+1 \right) n_B^{1 + \alpha} +
	\frac{9 t_3^{\prime}}{8} \left( \alpha+1 \right) n_B^{1 + \alpha}
	\f]
    */
    virtual double fcomp(double nb);

    /** 
	\brief Calculate symmetry energy

	If pf=0.5, then the exact expression below is used.
	Otherwise, the method from class hadronic_eos is used.

	\f[
	E_{sym} = \frac{5}{9} C n^{2/3} + \frac{10 C m}{3}
	\left[ \frac{t_2}{6} \left(1 + \frac{5}{4} x_2 \right) - 
	\frac{1}{8} t_1 x_1 \right] n^{5/3} 
	- \frac{t_3^{\prime}}{24} 
	\left({\textstyle \frac{1}{2}} + x_3 \right) n^{1+\alpha} - 
	\frac{t_0}{4} \left( {\textstyle \frac{1}{2}} + x_0 \right) n 
	\f]
    */
    virtual double fesym(const double nb, const double pf=0.5);

    /** 
	\brief skewness

	\f[
	2 C n_B^{2/3} \left(9-5/M^{*}/M\right)+
	\frac{27 t_3^{\prime}}{16} n^{1+\alpha} \alpha 
	\left(\alpha^2-1\right)
	\f]
    */
    virtual double fkprime(double nb);
    //@}

    /** \brief Calculate \f$ t_0,t_1,t_2,t_3 \f$ and \f$ \alpha \f$ from 
	the saturation properties.
      
	In nuclear matter: 
      
	\f$ E_b=E_b(n_0,M^{*},t_0,t_3,\alpha) \f$ \n
	\f$ P=P(n_0,M^{*},t_0,t_3,\alpha) \f$ \n
	\f$ K=K(n_0,M^{*},t_3,\alpha) \f$ 
	(the \f$ t_0 \f$ dependence vanishes) \n
	\f$ M^{*}=M^{*}(n_0,t_1,t_2,x_2) \f$ 
	(the \f$ x_1 \f$ dependence cancels), \n
	\f$ E_{sym}=E_{sym}(x_0,x_1,x_2,x_3,t_0,t_1,t_2,t_3,\alpha) \f$
      
	To fix the couplings from the saturation properties, we take
	\f$ n_0, M^{*}, E_b, K \f$ as inputs, and we can fix \f$
	t_0,t_3,\alpha \f$ from the first three relations, then use
	\f$ M^{*}, E_b \f$ to fix \f$ t_2 \f$ and \f$ t_1 \f$.  The
	separation into two solution steps should make for better
	convergence. All of the x's are free parameters and should be
	set before the function call.
      
	The arguments \c gt0, \c gt3, \c galpha, \c gt1, and \c gt2
	are used as initial guesses for ::t0, ::t3, ::alpha, ::t1, and
	::t2 respectively.
      
	\todo Does this work for both 'a' and 'b' non-zero?
      
	\todo Compare to similar formulae from \ref Margueron02
    */

    int calpar(double gt0=-10.0, double gt3=70.0, double galpha=0.2,
	       double gt1=2.0, double gt2=-1.0);

    // Unfinished.
    /* \brief 
	From \ref Margueron02
    */
    //  int calpar_new(double m);

    /** 
	\brief Load parameters from model 'model'

	Presently accepted values from file skdata/model_list:
	\include skdata/model_list
	
	If \c external is true, then model is the filename
	(relative to the current directory) of the file
	containing the model parameters
    */
    int load(std::string model, bool external=false);

    /** 
	\brief Use hadronic_eos methods for saturation properties
      
	This can be set to true to check the difference between
	the exact expressions and the numerical values from
	class hadronic_eos.
    */
    bool parent_method;
  
    /** 
	\brief Check the Landau parameters for instabilities

	This returns zero if there are no instabilities.
     */
    int check_landau(double nb, double m);

    /** 
	\brief Calculate the Landau parameters for nuclear matter

	Given \c n0 and \c m, this calculates the Landau parameters in
	nuclear matter as given in \ref Margueron02
     
	\todo This needs to be checked.
	
	(Checked once on 11/05/03)
    */
    int landau_nuclear(double n0, double m,
		       double &f0, double &g0, double &f0p,
		       double &g0p, double &f1, double &g1,
		       double &f1p, double &g1p);

    /** 
	\brief Calculate the Landau parameters for neutron matter
    
	Given 'n0' and 'm', this calculates the Landau parameters in
	neutron matter as given in \ref Margueron02
	
	\todo This needs to be checked
	
	(Checked once on 11/05/03)
    */
    int landau_neutron(double n0, double m, double &f0, double &g0, 
		       double &f1, double &g1);

    /// Default nonrelativistic neutron
    nonrel_fermion def_nr_neutron;

    /// Default nonrelativistic proton
    nonrel_fermion def_nr_proton;

    /// Return string denoting type ("skyrme_eos")
    virtual const char *type() { return "skyrme_eos"; }

#ifndef DOXYGENP
    
  protected:
    
    /// \name Functions and parameters for calpar()
    //@{
    int calparfun(size_t nv, const ovector_view &x, ovector_view &y, 
		  void *&vp);
    int calparfun2(size_t nv, const ovector_view &x, ovector_view &y, 
		   void *&vp);
    double fixn0, fixeoa, fixesym, fixcomp, fixmsom;
    //@}

#endif

  };

  template<> int io_tlate<skyrme_eos>::input
    (cinput *co, in_file_format *ins, skyrme_eos *sk);
  template<> int io_tlate<skyrme_eos>::output
    (coutput *co, out_file_format *outs, skyrme_eos *sk);
  template<> const char *io_tlate<skyrme_eos>::type();

  typedef io_tlate<skyrme_eos> skyrme_eos_io_type;

#ifndef DOXYGENP
}
#endif

#endif
