/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
 
#ifndef O2SCL_SYM4_EOS_H
#define O2SCL_SYM4_EOS_H

#include <iostream>
#include <o2scl/ovector_tlate.h>
#include <o2scl/apr_eos.h>
#include <o2scl/rmf_eos.h>
#include <o2scl/skyrme_eos.h>
#include <o2scl/gen_potential_eos.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A class to compute quartic contributions to the symmetry energy
   */
  class sym4_eos_base {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// An electron for the computation of the \f$ \hat{\mu}\f$
    fermion e;

#endif

  public:

    sym4_eos_base();
  
    virtual ~sym4_eos_base() {}

    /** \brief Compute alpha at the specified density
     */
    virtual int calc_e_alpha(fermion &ne, fermion &pr, thermo &lth,
			     double &alphak, double &alphap, double &alphat,
			     double &diff_kin, double &diff_pot,
			     double &ed_kin_nuc, double &ed_pot_nuc);
    
    /** \brief Compute \f$ \hat{\mu} \f$, the out-of-whack parameter
     */
    virtual double calc_muhat(fermion &ne, fermion &pr);

    /** \brief Compute the potential and kinetic parts separately (to
	be overwritten in children)
     */
    virtual int calc_e_sep(fermion &ne, fermion &pr, double &ed_kin, 
			   double &ed_pot, double &mu_n_kin, double &mu_p_kin, 
			   double &mu_n_pot, double &mu_p_pot);

  };

  /** \brief A version of \ref rmf_eos to separate potential and kinetic
      contributions
   */
  class rmf4_eos : public rmf_eos, public sym4_eos_base {

  public:
  
    rmf4_eos();
  
    /** \brief Compute the potential and kinetic parts separately
     */
    virtual int calc_e_sep(fermion &ne, fermion &pr, double &ed_kin, 
			   double &ed_pot, double &mu_n_kin, double &mu_p_kin, 
			   double &mu_n_pot, double &mu_p_pot);

  };

  /** \brief A version of \ref apr_eos to separate potential and kinetic
      contributions
   */
  class apr4_eos : public apr_eos, public sym4_eos_base {

  public:

    /** \brief Compute the potential and kinetic parts separately
     */
    virtual int calc_e_sep(fermion &ne, fermion &pr, double &ed_kin, 
			   double &ed_pot, double &mu_n_kin, double &mu_p_kin, 
			   double &mu_n_pot, double &mu_p_pot);
  };

  /** \brief A version of \ref skyrme_eos to separate potential and kinetic
      contributions
   */
  class skyrme4_eos : public skyrme_eos, public sym4_eos_base {

  public:

    /** \brief Compute the potential and kinetic parts separately
     */
    virtual int calc_e_sep(fermion &ne, fermion &pr, double &ed_kin, 
			   double &ed_pot, double &mu_n_kin, double &mu_p_kin, 
			   double &mu_n_pot, double &mu_p_pot);
  };
  
  /** \brief A version of \ref gen_potential_eos to 
      separate potential and kinetic contributions
   */
  class mdi4_eos : public gen_potential_eos, public sym4_eos_base {

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// Compute the kinetic part of the energy density
    double energy_kin(double var);

    /// Compute the potential part of the energy density
    double energy_pot(double var);

#endif
    
  public:

    /** \brief Compute the potential and kinetic parts separately
     */
    virtual int calc_e_sep(fermion &ne, fermion &pr, double &ed_kin, 
			   double &ed_pot, double &mu_n_kin, double &mu_p_kin, 
			   double &mu_n_pot, double &mu_p_pot);

    /// Test the separation of the potential and kinetic energy parts
    virtual int test_separation(fermion &ne, fermion &pr, test_mgr &t);

  };

  /** \brief Construct an EOS with an arbitrary choice for the terms
      in the symmetry energy that are quartic in the isospin asymmetry
  */
  class sym4_eos : public hadronic_eos {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// The base equation of state to use
    sym4_eos_base *sp;

#endif

  public:

    /// The strength of the quartic terms
    double alpha;

    /// Set the base equation of state
    int set_base_eos(sym4_eos_base &seb);
  
    /** 
	\brief Test the equation of state 
    
	This compares the chemical potentials from calc_e_sep() to
	their finite-difference approximations in order to ensure that
	the separation into potential and kinetic parts is done
	properly.
     */
    virtual int test_eos(fermion &ne, fermion &pr, thermo &lth);

    /** 
	\brief Equation of state as a function of density
    */
    virtual int calc_e(fermion &ne, fermion &pr, thermo &lth);

  };

#ifndef DOXYGENP
}
#endif

#endif
