/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/tov_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

#include <o2scl/test_mgr.h>
#include <o2scl/nonrel_fermion.h>
#include <o2scl/apr_eos.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/cern_mroot.h>

nonrel_fermion neut(939.0/hc_mev_fm,2.0),p(939.0/hc_mev_fm,2.0);
fermion e(0.511/hc_mev_fm,2.0),zero(0.0,0.0);
thermo ht, hb, l;
apr_eos ap;
double barn;
int nstarfun(size_t nv, const ovector_view &x, ovector_view &y, void *&vp);

const double xmev_kg=1.782661731e-30; 
const double mev_per_fm3_msun_km3=xmev_kg/gsl_mks::solar_mass*1.0e54; 

int main(void) {
  test_mgr t;
  t.set_output_level(2);
  gsl_mroot_hybrids<void *,mm_funct<void *> > nd;
  cern_mroot<void *,mm_funct<void *> > nd2;
  ovector x(4);
  void *vp=0;
  table eos(200);
  collection co;

  cout.setf(ios::scientific);

  // -----------------------------------------------------
  // Calculate APR EOS - 
  // Avoid the complication of a Gibbs construction
  // for now.

  ap.select(1);
  ap.pion=0;

  x[0]=0.09663;
  x[1]=0.003365;
  x[2]=0.4636;

  eos.line_of_names("ed pr nb");
  mm_funct_fptr<void *> nf(nstarfun);
  for(barn=0.04;barn<=2.0001;barn+=0.01) {
    int ret=nd.msolve(3,x,vp,nf);
    // If the GSL solver fails, use CERNLIB
    if (ret!=0) {
      int ret2=nd2.msolve(3,x,vp,nf);
      if (ret2!=0) {
	cout << "Solver failed." << endl;
      }
    }
    double line[3]={ht.ed*mev_per_fm3_msun_km3*hc_mev_fm,
		    ht.pr*mev_per_fm3_msun_km3*hc_mev_fm,
		    barn};
    eos.line_of_data(3,line);
  }
  
  tov_interp_eos te;
  te.verbose=3;
  te.set_low_density_eos(true,"");
  te.read_table(eos);
  
  double ed, nb;
  t.test_gen(te.baryon_column==true,"baryon column");
  vector<string> auxp;
  size_t np;
  te.get_names(np,auxp);
  t.test_gen(np==0,"no aux parms");
  
  cout << "Switching to units of inverse fm." << endl;
  for(size_t ik=0;ik<eos.get_nlines();ik++) {
    eos.set("ed",ik,eos.get("ed",ik)/(mev_per_fm3_msun_km3*hc_mev_fm));
    eos.set("pr",ik,eos.get("pr",ik)/(mev_per_fm3_msun_km3*hc_mev_fm));
  }
  te.set_units("fm^-4","fm^-4","fm^-3");
  te.read_table(eos);
  
  cout << endl;
  cout << "User (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden_user(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }
  cout << endl;
  cout << "LD (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden_ld(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }
  cout << endl;
  cout << "Mix (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }

  double p1, p2, p3;
  te.get_transition(p1,p2,p3);
  cout << endl;
  cout << p1 << " " << p2 << " " << p3 << endl;
  
  te.set_transition(2.0e-3,1.2);

  cout << endl;
  cout << "User (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden_user(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }
  cout << endl;
  cout << "LD (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden_ld(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }
  cout << endl;
  cout << "Mix (e,p,nb): " << endl;
  for (double P2=1.5e-3;P2<=3.0e-3;P2*=1.07) {
    te.get_eden(P2,ed,nb);
    cout << ed << " " << P2 << " " << nb << endl;
  }

  te.get_transition(p1,p2,p3);
  cout << endl;
  cout << p1 << " " << p2 << " " << p3 << endl;
  
  t.report();

  return 0;
}

int nstarfun(size_t nv, const ovector_view &x, ovector_view &y, void *&vp) {
  neut.n=x[0];
  p.n=x[1];
  e.mu=x[2];

  if (x[0]<0.0 || x[1]<0.0) return 1;

  ap.calc_e(neut,p,hb);
  e.calc_mu_zerot();

  l.ed=e.ed;
  l.pr=e.pr;
  ht=l+hb;

  y[0]=neut.n+p.n-barn;
  y[1]=p.n-e.n;
  y[2]=neut.mu-p.mu-e.mu;

  return 0;
}


