/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_GSL_MIN_BRENT_H
#define O2SCL_GSL_MIN_BRENT_H

#include <gsl/gsl_min.h>
#include <o2scl/minimize.h>
 
#ifndef DOXYGENP
namespace o2scl {
#endif
   
  /** 
      \brief One-dimensional minimization (GSL)

      \todo Simplify temporary storage and document 
      stopping conditions.
  */
  template<class param_t, class func_t=funct<param_t> > class gsl_min_brent : 
  public minimize<param_t,func_t> 
    {
	
#ifndef DOXYGEN_INTERNAL
	
    protected:

      /// The function
      func_t *uf;
      
      /// The parameters
      param_t *up;
      
      /// \name Temporary storage
      //@{
      double d, e, v, w, f_v, f_w;
      //@}
  
      /// Compute the function values at the various points
      int compute_f_values(func_t &func, double xminimum, double *fminimum, 
			   double xlower, double *flower, double xupper,
			   double *fupper, param_t &pa) {
    
	func(xlower,*flower,pa);
	func(xupper,*fupper,pa);
	func(xminimum,*fminimum,pa);
    
	return gsl_success;
      }
      
#endif
      
    public:
 
      /// Location of minimum
      double x_minimum;
      /// Lower bound
      double x_lower;
      /// Upper mound
      double x_upper;
      /// Minimum value
      double f_minimum;
      /// Value at lower bound
      double f_lower;
      /// Value at upper bound
      double f_upper;

      /// Set the function to be minimized and the initial brackeing interval
      int set(func_t &func, double xmin, double lower, double upper,
	      param_t &pa) {
      
	double fmin, fl, fu;
	int status=compute_f_values(func,lower,&fl,xmin,&fmin,upper,&fu,pa);
    
	status=set_with_values(func,xmin,fmin,lower,fl,upper,fu,pa);

	return status;
      }

      /// Set the function to be minimized and the initial brackeing interval
      int set_with_values(func_t &func, double xmin, double fmin,
			  double lower, double fl, double upper,
			  double fu, param_t &pa) {

	if (lower > upper) {
	  set_err_ret("Invalid interval in set_with_values().",gsl_einval);
	}
	if (xmin>=upper || xmin<=lower) {
	  set_err_ret
	    ("'xmin' was not inside interval set_with_values().",gsl_einval);
	}
	if (fmin >= fl || fmin>=fu) {
	  set_err_ret
	    ("Endpoints don't enclose minimum in set_with_values().",
	     gsl_einval);
	}
    
	x_lower=lower;
	x_minimum=xmin;
	x_upper=upper;
	f_lower=fl;
	f_minimum=fmin;
	f_upper=fu;

	uf=&func;
	up=&pa;

	/* golden=(3-sqrt(5))/2 */
	const double golden=0.3819660;      
    
	v=x_lower+golden*(x_upper-x_lower);
	w=v;
	d=0;
	e=0;
	func(v,f_v,pa);
	f_w=f_v;
    
	return gsl_success;
      }
  
      /// Perform an iteration
      int iterate() {
    
	const double x_left= x_lower;
	const double x_right= x_upper;
	
	const double z=x_minimum;
	double u, f_u;
	const double f_z=f_minimum;
	
	/* golden=(3-sqrt(5))/2 */
	const double golden=0.3819660;      
    
	const double w_lower=(z-x_left);
	const double w_upper=(x_right-z);
    
	const double tolerance= GSL_SQRT_DBL_EPSILON*fabs (z);
    
	double p=0, q=0, r=0;
    
	const double midpoint=0.5*(x_left+x_right);
    
	if (fabs (e) > tolerance) {
	  /* fit parabola */
	  
	  r=(z-w)*(f_z-f_v);
	  q=(z-v)*(f_z-f_w);
	  p=(z-v)*q-(z-w)*r;
	  q=2*(q-r);
	  
	  if (q > 0) {
	    p=-p;
	  } else {
	    q=-q;
	  }
	  
	  r=e;
	  e=d;
	}

	if (fabs (p) < fabs (0.5*q*r) && p < q*w_lower && 
	    p < q*w_upper) {

	  double t2=2*tolerance;
	  
	  d=p / q;
	  u=z+d;
	  
	  if ((u-x_left) < t2 || (x_right-u) < t2)
	    {
	      d=(z < midpoint) ? tolerance : -tolerance ;
	    }
	} else {
	  e=(z < midpoint) ? x_right-z : -(z-x_left) ;
	  d=golden*e;
	}
    
    
	if (fabs (d) >= tolerance) {
	  u=z+d;
	} else {
	  u=z+((d > 0) ? tolerance : -tolerance) ;
	}
	
	(*uf)(u,f_u,*up);
	
	if (f_u > f_z) {

	  if (u < z) {
	    x_lower=u;
	    f_lower=f_u;
	    return gsl_success;
	  } else {
	    x_upper=u;
	    f_upper=f_u;
	    return gsl_success;
	  }

	} else if (f_u < f_z) {

	  if (u < z) {
	    x_upper=z;
	    f_upper=f_z;
	  } else {
	    x_lower=z;
	    f_lower=f_z;
	  }

	  v=w;
	  w=z;
	  f_v=f_w;
	  f_w=f_z;
	  x_minimum=u;
	  f_minimum=f_u;

	  return gsl_success;

	} else if (f_u <= f_w || w == z) {

	  v=w;
	  f_v=f_w;
	  w=u;
	  f_w=f_u;
	  return gsl_success;

	} else if (f_u <= f_v || v == z || v == w) {

	  v=u;
	  f_v=f_u;
	  return gsl_success;

	} else {

	  set_err_ret("Iteration failed in gsl_min_brent::iterate().",
		      gsl_failure);

	}
	
	set_err("Sanity check failed in gsl_min_brent::iterate().",
		gsl_esanity);
	return gsl_esanity;
      }
  
      virtual ~gsl_min_brent() {}
  
      /** \brief Calculate the minimum \c min of \c func 
	  with \c x2 bracketed between \c x1 and \c x3.

	  Note that this algorithm requires that the initial guess
	  already brackets the minimum, i.e. \f$ x_1 < x_2 < x_3 \f$,
	  \f$ f(x_1) > f(x_2) \f$ and \f$ f(x_3) > f(x_2) \f$. This is
	  different from \ref cern_minimize, where the initial value
	  of the first parameter to cern_minimize::min_bkt() is
	  ignored.
      */
      virtual int min_bkt(double &x2, double x1, double x3, double &fmin,
			  param_t &pa, func_t &func) {
    
	int status, iter=0;
	double a,b;
	double ytmp;
    
	set(func,x2,x1,x3,pa);
    
	do {
	  iter++;
	  status=iterate();
	  if (status) break;
      
	  a=x_lower;
	  b=x_upper;
	  x2=x_minimum;
      
	  status=gsl_min_test_interval(a,b,this->tolx,this->tolf);
      
	  if (this->verbose>0) {
	    func(x2,ytmp,pa);
	    if (a*b<0.0) {
	      if (a<b) {
		print_iter(x2,ytmp,iter,fabs(a-b)-this->tolf*a,this->tolx);
	      } else {
		print_iter(x2,ytmp,iter,fabs(a-b)-this->tolf*b,this->tolx);
	      }
	    } else {
	      print_iter(x2,ytmp,iter,fabs(a-b),this->tolx);
	    }
	  }
      
	} while (status == gsl_continue && iter<this->ntrial);
	
	this->last_ntrial=iter;
	func(x2,fmin,pa);
    
	if (status==gsl_success) return 0;
    
	return status;
      }
  
      /// Return string denoting type ("gsl_min_brent")
      virtual const char *type() { return "gsl_min_brent"; }
      
    };

#ifndef DOXYGENP
}
#endif

#endif
