/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MINIMIZE_H
#define O2SCL_MINIMIZE_H
 
/** \file minimize.h
    \brief One-dimensional minimization routines
*/


#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Numerical differentiation base
  */
  template<class param_t, class func_t, class dfunc_t=func_t> 
    class minimize {
    
    public:
    
    /// Output control
    int verbose;
    
    /// Maximum number of iterations
    int ntrial;
    
    /// The tolerance for the minimum function value
    double tolf;

    /// The tolerance for the location of the minimum
    double tolx;
    
    /// The number of iterations for in the most recent minimization
    int last_ntrial;
    
    minimize() {
      verbose=0;
      ntrial=100;
      tolf=1.0e-4;
      tolx=1.0e-4;
      last_ntrial=0;
    }
      
    virtual ~minimize() {}
      
    /** 
	\brief Print out iteration information.
	
	Depending on the value of the variable verbose, this prints out
	the iteration information. If verbose=0, then no information is
	printed, while if verbose>1, then after each iteration, the
	present values of x and y are output to std::cout along with the
	iteration number. If verbose>=2 then each iteration waits for a
	character.  
    */
    virtual int print_iter(double x, double y, int iter, double value=0.0,
			   double limit=0.0, std::string comment="") {

      if (verbose<=0) return gsl_success;
    
      char ch;
    
      std::cout << comment << " Iteration: " << iter << std::endl;
      if (x<0) std::cout << x << " ";
      else std::cout << " " << x << " ";
      if (y<0) std::cout << y << " ";
      else std::cout << " " << y << " ";
      if (value<0) std::cout << value << " ";
      else std::cout << " " << value << " ";
      if (limit<0) std::cout << limit << std::endl;
      else std::cout << " " << limit << std::endl;
      if (verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }
    
      return gsl_success;
    }
  
    /** 
	\brief Calculate the minimum \c min of \c func w.r.t 'x'.

	If this is not overloaded, it attempts to bracket the 
	minimum using bracket() and then calls min_bkt() with
	the newly bracketed minimum.
    */
    virtual int min(double &x, double &fmin, param_t &pa, func_t &func) {
      double xl, xr, f, fl, fr;
      xl=x*0.9;
      xr=x*1.1;
      bracket(xl,x,xr,fl,f,fr,pa,func);
      return min_bkt(x,xl,xr,fmin,pa,func);
    }
    
    /** \brief Calculate the minimum \c min of \c func with x2
	bracketed between x1 and x3
	  
	If this is not overloaded, it ignores the bracket and calls min().  
    */
    virtual int min_bkt(double &x2, double x1, double x3, double &fmin,
			param_t &pa, func_t &func) 
    {
      return min(x2,fmin,pa,func);
    }
    
    /** \brief Calculate the minimum \c min of \c func with
	derivative \c dfunc w.r.t 'x'.
	
	If this is not overloaded, it attempts to bracket the 
	minimum using bracket() and then calls min_bkt_de() with
	the newly bracketed minimum.
    */
    virtual int min_de(double &x, double &fmin, param_t &pa, func_t &func,
		       dfunc_t &df) 
    {
      double xl, xr, f, fl, fr;
      xl=x*0.9;
      xr=x*1.1;
      bracket(xl,x,xr,fl,f,fr,pa,func);
      return min_bkt_de(x,xl,xr,fmin,pa,func,df);
    }
    
    /** \brief Calculate the minimum \c min of \c func with derivative
	\c dfunc and x2 bracketed between x1 and x3
	
	If this is not overloaded, it ignores the bracket and calls min_de().
    */
    virtual int min_bkt_de(double &x2, double x1, double x3, double &fmin, 
			   param_t &pa, func_t &func, dfunc_t &df) 
    {
      return min_de(x2,fmin,pa,func,df);
    }
  
    /** 
	\brief Given ax and bx, bracket a minimum for function \c func.
	
	Upon success, fa=f(ax), fb=f(bx), and fc=f(cx) with 
	fb<fa and fb<fc and ax<bx<cx. 
	
	\todo Improve this algorithm with the standard golden ratio
	method.
	\todo Double check that this works when at least two of 
	f(a), f(b) and f((a+b)/2) are equal.
    */
    virtual int bracket(double &ax, double &bx, double &cx, double &fa, 
			double &fb, double &fc, param_t &pa, 
			func_t &func) 
    {
			    
      double x=ax, x2=bx, x3=(ax+bx)/2.0;
      double fx, fx3, fx2;
      int i=0;
    
      bool done=false;
      while(done==false && i<20) {
	func(x,fx,pa);
	func(x2,fx2,pa);
	func(x3,fx3,pa);
      
	if (fx3>=fx2 && fx3>=fx) {
	  if (fx2>fx) {
	    x2=x3;
	    x3=(x+x2)/2.0;
	  } else {
	    x=x3;
	    x3=(x+x2)/2.0;
	  }
	} else if (fx<=fx3 && fx3<=fx2) {
	  x3=x;
	  x=x2-2.0*(x2-x);
	} else if (fx3<fx2 && fx3<fx) {
	  done=true;
	  ax=x;
	  bx=x3;
	  cx=x2;
	  fa=fx;
	  fb=fx3;
	  fc=fx2;
	} else {
	  x3=x2;
	  x2=x+2.0*(x2-x);
	}
	i++;
      }
    
      if (done==false) {
	set_err_ret("Too many iterations in minimize::bracket().",
		    gsl_emaxiter);
      }
    
      return 0;
    }
  
    /// Return string denoting type ("minimize")
    virtual const char *type() { return "minimize"; }
  
  };

  /** 
      \brief Constrain \c x to be within \c width
      of the value given by \c center
      
      Defining \f$ c= \f$ \c center, \f$ w= \f$ \c width, \f$ h= \f$
      \c height, this returns the value \f$ h (1+|x-c-w|/w) \f$ if \f$
      x>c+w \f$ and \f$ h (1+|x-c+w|/w) \f$ if \f$ x<c-w \f$ . The
      value at \f$ x=c-w \f$ or \f$ x=c+w \f$ is \f$ h \f$ and the
      value at \f$ x=c-2w \f$ or \f$ x=c+2w \f$ is \f$ 2 h \f$ .

      It is important to note that, for large distances of \c x 
      from \c center, this only scales linearly. If you are trying to
      constrain a function which decreases more than linearly by
      making \c x far from \c center, then a minimizer will
      likely ignore this constraint.
  */
  inline double constraint(double x, double center, double width, 
			   double height) {
    double ret=0.0;
    if (x>center+width) {
      ret=height*(1.0+fabs(x-center-width)/width);
    } else if (x<center-width) {
      ret=height*(1.0+fabs(x-center+width)/width);
    }
    return ret;
  }

  /** \brief Constrain \c x to be within \c width of the value given
      by \c center

      Defining \f$ c= \f$ \c center, \f$ w= \f$ \c width, \f$ h= \f$
      \c height, \f$ t= \f$ \c tightness, and \f$ \ell= \f$ \c
      exp_arg_limit, this returns the value
      \f[
      \left(\frac{x-c}{w}\right)^2 \left[ 
      1+ e^{t\left(x-c-w\right)\left(c+w-x\right)/w^2}
      \right]^{-1}
      \f]
      if \f$ x \geq c \f$ .

      The exponential is handled gracefully by assuming that anything
      smaller than \f$ \exp(-\ell) \f$ is zero. This function is
      continuous and differentiable.  Note that if \f$ x=c \f$ , then
      the function returns zero.
      
      It is important to note that, for large distances of \c x 
      from \c center, this only scales linearly. If you are trying to
      constrain a function which decreases more than linearly by
      making \c x far from \c center, then a minimizer will
      likely ignore this constraint.
  */
  inline double cont_constraint(double x, double center, double width, 
				double height, double tightness=40.0, 
				double exp_arg_limit=50.0) {
    double ret, wid2=width*width;
    double arg=tightness/wid2*(x-center-width)*(center+width-x);
    if (arg<-exp_arg_limit) {
      ret=(x-center)*(x-center)/wid2;
    } else {
      ret=(x-center)*(x-center)/wid2/(1.0+exp(arg));
    }
    return ret;
  }
  
  /** 
      \brief Constrain \c x to be greater than the value given by \c
      center

      Defining \f$ c= \f$ \c center, \f$ w= \f$ \c width, \f$ h= \f$
      \c height, this returns \f$ h(1+|x-c|/w) \f$ if \f$ x<c \f$ and
      zero otherwise.  The value at \f$ x=c \f$ is \f$ h \f$ , while
      the value at \f$ x=c-w \f$ is \f$ 2 h \f$ .

      It is important to note that, for large distances of \c x 
      from \c center, this only scales linearly. If you are trying to
      constrain a function which decreases more than linearly by
      making \c x far from \c center, then a minimizer will
      likely ignore this constraint.
  */
  inline double lower_bound(double x, double center, double width, 
			    double height) {
    double ret=0.0;
    if (x<center) ret=height*(1.0+fabs(x-center)/width);
    return ret;
  }
  
  /** 
      \brief Constrain \c x to be greater than the value given by \c
      center 

      Defining \f$ c= \f$ \c center, \f$ w= \f$ \c width, \f$ h= \f$ \c
      height, \f$ t= \f$ \c tightness, and \f$ \ell= \f$ \c exp_arg_limit,
      this returns \f$ h(c-x+w)/(1+\exp(t(x-c)/w)) \f$ and has the
      advantage of being a continuous and differentiable function. The
      exponential is handled gracefully by assuming that anything
      smaller than \f$ \exp(-\ell) \f$ is zero

      It is important to note that, for large distances of \c x 
      from \c center, this only scales linearly. If you are trying to
      constrain a function which decreases more than linearly by
      making \c x far from \c center, then the constraint will 
      be essentially ignored.
  */
  inline double cont_lower_bound(double x, double center, double width, 
				 double height, double tightness=40.0, 
				 double exp_arg_limit=50.0) {
    double ret, arg=tightness*(x-center)/width;
    if (arg>exp_arg_limit) {
      ret=0.0;
    } else if (arg<-exp_arg_limit) {
      ret=height*(center-x+width);
    } else {
      ret=height*(center-x+width)/(1.0+exp(arg));
    }
    return ret;
  }

#ifndef DOXYGENP
}
#endif

#endif
