/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#include <o2scl/nuclear_mass.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;
using namespace o2scl_fm;

nuclear_mass::nuclear_mass() {

  std::string tlist[119]=
    {"n","H","He","Li","Be","B","C","N","O","F","Ne","Na","Mg","Al","Si",
     "P","S","Cl","Ar","K","Ca","Sc","Ti","V","Cr","Mn","Fe","Co","Ni",
     "Cu","Zn","Ga","Ge","As","Se","Br","Kr","Rb","Sr","Y","Zr","Nb","Mo",
     "Tc","Ru","Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe","Cs","Ba",
     "La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy","Ho","Er","Tm",
     "Yb","Lu","Hf","Ta","W","Re","Os","Ir","Pt","Au","Hg","Tl","Pb","Bi",
     "Po","At","Rn","Fr","Ra","Ac","Th","Pa","U","Np","Pu","Am","Cm","Bk",
     "Cf","Es","Fm","Md","No","Lr","Rf","Db","Sg","Bh","Hs","Mt","Ds","Rg",
     "Uub","Uut","Uuq","Uup","Uuh","Uus","Uuo"};
  
  for(int i=0;i<nelements;i++) {
    element_list[i]=tlist[i];
    element_table.insert(make_pair(element_list[i],i));
  }

  form_type=0;
}

mnms95_mass::mnms95_mass() {
  form_type=disc_type;

  void *vp;
  int sz, sz2;

  string file;
  collection co;
  
  base_ioc bio;
  mnms95_mass_entry_io_type mio;

  file_detect fd;
  file=lib_settings.get_data_dir()+"/nndc/moller.dat";
  
  in_file_format *iff=fd.open(file.c_str());
  
  if (co.fin(iff)==0) {
    
    string mm_name;
    /// Need to check error condition for this part
    collection::type_iterator it=co.begin("mnms95_mass_entry");
    mm_name=it.name();
    sz=it->size;
    sz2=it->size2;
    vp=it->data;
    
    n=sz;
    mass=(mnms95_mass_entry *)vp;
    
    co.disown(mm_name);
    co.clear();
    
    loaded=true;
    
  } else {

    std::string err="Failed to load file "+file+
      " in mnms95_mass::mnms95_mass().";
    set_err(err.c_str(),gsl_efilenotfound);
    loaded=false;
    
  }
  
  fd.close();
  
  last=n/2;

  return;
}

bool mnms95_mass::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    return true;
  }


  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      return true;
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }
  }
  
  return false;
}

bool mnms95_mass_exp::is_included(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;

  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      return false;
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    if (fabs(mass[mid].Mexp)>1.0e-20) {
      return true;
    } else {
      return false;
    }
  }


  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      if (fabs(mass[mid].Mexp)>1.0e-20) {
	return true;
      } else {
	return false;
      }
    } else if (mass[mid].N>l_N) {
      if (mid==0) return false;
      mid--;
    } else {
      if (mid==n-1) return false;
      mid++;
    }
  }
  
  return false;
}

mnms95_mass_entry mnms95_mass::get_ZN(int l_Z, int l_N) {
  int lo=0, hi=0, mid=last;
  
  mnms95_mass_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  
  // binary search for the correct Z first
  if (mass[mid].Z!=l_Z) {
    if (mass[mid].Z>l_Z) {
      lo=0;
      hi=mid;
    } else {
      lo=mid;
      hi=n-1;
    }
    while (hi>lo+1) {
      int mp=(lo+hi)/2;
      if (mass[mp].Z<l_Z) {
	lo=mp;
      } else {
	hi=mp;
      }
    }
    mid=lo;
    if (mass[mid].Z!=l_Z) mid=hi;
    if (mass[mid].Z!=l_Z) {
      set_err((((string)"Nuclei with Z=")+itos(l_Z) 
	       +" not found in mnms95_mass::get_ZN().").c_str(),
	      gsl_notfound);
    }
  }

  // The cached point was the right one, so we're done
  if (mass[mid].N==l_N) {
    ret=mass[mid];
    last=mid;
    return ret;
  }

  // Now look for the right N among all the Z's
  while (mass[mid].Z==l_Z) {
    if (mass[mid].N==l_N) {
      ret=mass[mid];
      last=mid;
      return ret;
    } else if (mass[mid].N>l_N) {
      mid--;
    } else {
      mid++;
    }
  }

  set_err((((string)"Nuclei with Z=")+itos(l_Z) 
	   +" not found in mnms95_mass::get_ZN().").c_str(),gsl_notfound);
  return ret;
}

template<> int io_tlate<mnms95_mass_entry>::input
(cinput *co, in_file_format *ins, mnms95_mass_entry *t) {
  ins->int_in(t->N);
  ins->int_in(t->A);
  t->Z=t->A-t->N;
  ins->double_in(t->eps2);
  ins->double_in(t->eps3);
  ins->double_in(t->eps4);
  ins->double_in(t->eps6);
  ins->double_in(t->eps6sym);
  ins->double_in(t->beta2);
  ins->double_in(t->beta3);
  ins->double_in(t->beta4);
  ins->double_in(t->beta6);
  ins->double_in(t->Emic);
  ins->double_in(t->Mth);
  ins->double_in(t->Mexp);
  ins->double_in(t->sigmaexp);
  ins->double_in(t->EmicFL);
  ins->double_in(t->MthFL);
  return 0;
}

template<> int io_tlate<mnms95_mass_entry>::output
(coutput *co, out_file_format *outs, mnms95_mass_entry *t) {
  outs->int_out(t->N);
  outs->int_out(t->A);
  outs->double_out(t->eps2);
  outs->double_out(t->eps3);
  outs->double_out(t->eps4);
  outs->double_out(t->eps6);
  outs->double_out(t->eps6sym);
  outs->double_out(t->beta2);
  outs->double_out(t->beta3);
  outs->double_out(t->beta4);
  outs->double_out(t->beta6);
  outs->double_out(t->Emic);
  outs->double_out(t->Mth);
  outs->double_out(t->Mexp);
  outs->double_out(t->sigmaexp);
  outs->double_out(t->EmicFL);
  outs->double_out(t->MthFL);
  return 0;
}

template<> const char *io_tlate<mnms95_mass_entry>::type()
{ return "mnms95_mass_entry"; }
  
ame_mass::ame_mass(std::string version) {
  form_type=disc_type;

  void *vp;
  int sz, sz2;

  string file;
  collection co;

  base_ioc bio;
  ame_entry95_io_type ame_entry95_io;
  ame_entry03_io_type ame_entry03_io;

  file_detect fd;
  if (version=="95exp") {
    file=lib_settings.get_data_dir()+"/nndc/ame95exp.dat";
  } else if (version=="95rmd") {
    file=lib_settings.get_data_dir()+"/nndc/ame95rmd.dat";
  } else if (version=="03round") {
    file=lib_settings.get_data_dir()+"/nndc/ame03round.dat";
  } else {
    file=lib_settings.get_data_dir()+"/nndc/ame03.dat";
  }
  in_file_format *iff=fd.open(file.c_str());

  if (co.fin(iff)==0) {

    string *stemp;
    
    string_io_type string_io;
    string_io.get(co,"reference",stemp);
    reference=*stemp;
    
    co.get("short_names",vp,sz);
    short_names=(string *)vp;
    co.get("col_names",vp,sz);
    col_names=(string *)vp;
    
    string ame_name;
    if (version=="95exp" || version=="95rmd") {
      collection::type_iterator it=co.begin("ame_entry95");
      ame_name=it.name();
      sz=it->size;
      sz2=it->size2;
      vp=it->data;
    } else {
      collection::type_iterator it=co.begin("ame_entry03");
      ame_name=it.name();
      sz=it->size;
      sz2=it->size2;
      vp=it->data;
    }
    
    n=sz;
    mass=(ame_entry *)vp;
    
    co.disown("short_names");
    co.disown("col_names");
    co.disown(ame_name);
    co.clear();
    loaded=true;
    
  } else {
    
    std::string err="Failed to load file "+file+" in ame_mass::ame_mass().";
    set_err(err.c_str(),gsl_efilenotfound);
    
    loaded=false;
    
  }

  fd.close();
  
  
  return;
}

bool ame_mass::is_included(int l_Z, int l_N) {
  int i;
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  for(i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].N==l_N) {
      return true;
    }
  }
  return false;
}

ame_entry ame_mass::get_ZN(int l_Z, int l_N) {
  int i;
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  for(i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].N==l_N) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get_ZA(int l_Z, int l_A) {
  int i;
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  for(i=0;i<n;i++) {
    if (mass[i].Z==l_Z && mass[i].A==l_A) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get_elA(string l_el, int l_A) {
  int i;
  ame_entry ret;
  ret.Z=0;
  ret.A=0;
  ret.N=0;
  for(i=0;i<n;i++) {
    if (mass[i].el==l_el && mass[i].A==l_A) {
      ret=mass[i];
    }
  }
  return ret;
}

ame_entry ame_mass::get(string nucleus) {
  if (isalpha(nucleus[1])) {
    string el=nucleus.substr(0,2);
    int A=stoi(nucleus.substr(2,nucleus.length()-2));
    return get_elA(el,A);
  } 
  string el=nucleus.substr(0,1);
  int A=stoi(nucleus.substr(1,nucleus.length()-1));
  return get_elA(el,A);
}

int ame_entry95_io_type::input(cinput *co, in_file_format *ins, ame_entry *t) {
  ins->int_in(t->NMZ);
  ins->int_in(t->N);
  ins->int_in(t->Z);
  ins->int_in(t->A);
  ins->word_in(t->el);
  ins->word_in(t->orig);
  ins->double_in(t->mass);
  ins->double_in(t->dmass);
  ins->double_in(t->be);
  ins->double_in(t->dbe);
  t->beoa=t->be/((double)t->A);
  t->dbeoa=t->dbe/((double)t->A);
  ins->word_in(t->bdmode);
  ins->double_in(t->bde);
  ins->double_in(t->dbde);
  ins->int_in(t->A2);
  ins->double_in(t->amass);
  ins->double_in(t->damass);
  return 0;
}

int ame_entry95_io_type::output(coutput *co, out_file_format *outs, 
				ame_entry *t) {
  outs->int_out(t->NMZ);
  outs->int_out(t->N);
  outs->int_out(t->Z);
  outs->int_out(t->A);
  outs->word_out(t->el);
  outs->word_out(t->orig);
  outs->double_out(t->mass);
  outs->double_out(t->dmass);
  outs->double_out(t->be);
  outs->double_out(t->dbe);
  outs->word_out(t->bdmode);
  outs->double_out(t->bde);
  outs->double_out(t->dbde);
  outs->int_out(t->A2);
  outs->double_out(t->amass);
  outs->double_out(t->damass);
  return 0;
}

const char *ame_entry95_io_type::type() { return "ame_entry95"; }
  
int ame_entry03_io_type::input(cinput *co, in_file_format *ins, ame_entry *t) {
  ins->int_in(t->NMZ);
  ins->int_in(t->N);
  ins->int_in(t->Z);
  ins->int_in(t->A);
  ins->word_in(t->el);
  ins->word_in(t->orig);
  ins->double_in(t->mass);
  ins->double_in(t->dmass);
  ins->double_in(t->beoa);
  ins->double_in(t->dbeoa);
  t->be=t->beoa*((double)t->A);
  t->dbe=t->dbeoa*((double)t->A);
  ins->word_in(t->bdmode);
  ins->double_in(t->bde);
  ins->double_in(t->dbde);
  ins->int_in(t->A2);
  ins->double_in(t->amass);
  ins->double_in(t->damass);
  return 0;
}

int ame_entry03_io_type::output(coutput *co, out_file_format *outs, 
				ame_entry *t) {
  outs->int_out(t->NMZ);
  outs->int_out(t->N);
  outs->int_out(t->Z);
  outs->int_out(t->A);
  outs->word_out(t->el);
  outs->word_out(t->orig);
  outs->double_out(t->mass);
  outs->double_out(t->dmass);
  outs->double_out(t->beoa);
  outs->double_out(t->dbeoa);
  outs->word_out(t->bdmode);
  outs->double_out(t->bde);
  outs->double_out(t->dbde);
  outs->int_out(t->A2);
  outs->double_out(t->amass);
  outs->double_out(t->damass);
  return 0;
}

const char *ame_entry03_io_type::type() { return "ame_entry03"; }

