/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NUCLEAR_MASS_H
#define O2SCL_NUCLEAR_MASS_H

#include <cmath>
#include <string>
#include <map>
#include <o2scl/nucleus.h>
#include <o2scl/constants.h>
#include <o2scl/base_ioc.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Nuclear mass formula base

      \note Some mass formulas are undefined for sufficiently exotic 
      nuclei. Use the function is_included() to find if a particular
      nucleus is included or not.

      \note All descendants ought to set \ref form_type to indicate
      whether the mass formula is discrete or continuous
      
      Elements 112-118 are named "Uub", "Uut", "Uuq", "Uup",
      "Uuh", "Uus", and "Uuo", respectively.
      
      The binding energy is defined
      \f[
      \mathrm{BE}=Z m_{H} + N m_{n} - m_{\mathrm{nuclide}}
      \f]
      where \f$m_{\mathrm{nuclide}}\f$ is the mass of the
      nucleus including the mass of the electrons.
      The mass excess is defined
      \f[
      m_{\mathrm{excess}} = m_{\mathrm{nuclide}} - A m_{u}
      \f]
      
      For example, for \f$ \mathrm{U}^{238} \f$, the binding energy is
      1801.695 MeV, the mass excess is 47.30366 MeV, and
      \f$m_{\mathrm{nuclide}}\f$ is 221742.9 MeV. This is consistent
      with the above, as \f$m_H\f$ is 938.7830 MeV, \f$m_n\f$ is
      939.5650 MeV, and \f$m_u\f$ is 931.494 MeV.

      Some common reaction Q-values and separation energies: \n
      \f$\mathrm{Q}(\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+1)}\f$: 
      Beta-decay energy \n
      \f$\mathrm{Q}(2\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+2)}\f$: 
      Double beta-decay energy \n
      \f$\mathrm{Q}(4\beta^{-})=\mathrm{M(A,Z)}-\mathrm{M(A,Z+4)}\f$: 
      Four beta-decay energy \n
      \f$\mathrm{Q}(\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-4,Z-2)}-
      \mathrm{M(He^{4})}\f$: 
      Alpha-decay energy \n
      \f$\mathrm{Q}(\beta-n)=\mathrm{M(A,Z)}-\mathrm{M(A-1,Z+1)}
      -\mathrm{M(n)}\f$: 
      Beta-delayed neutron emission decay energy \n
      \f$\mathrm{Q}(d,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-2,Z-1)}
      -\mathrm{M(He^{4})}-\mathrm{M(H^{2})}\f$: 
      \f$(d,\alpha)\f$ reaction energy \n
      \f$\mathrm{Q}(\mathrm{EC})=\mathrm{M(A,Z)}-\mathrm{M(A,Z-1)}\f$: 
      Electron capture decay energy \n
      \f$\mathrm{Q}(\mathrm{ECp})=\mathrm{M(A,Z)}-\mathrm{M(A-1,Z-2)}\f$: 
      Electron capture with delayed proton emission decay energy \n
      \f$\mathrm{Q}(n,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-3,Z-2)}
      -\mathrm{M(He^{4})}+\mathrm{M(n)}\f$: 
      \f$(n,\alpha)\f$ reaction energy \n
      \f$\mathrm{Q}(p,\alpha)=\mathrm{M(A,Z)}-\mathrm{M(A-3,Z-1)}
      -\mathrm{M(He^{4})}+\mathrm{M(p)}\f$: 
      \f$(p,\alpha)\f$ reaction energy \n
      \f$\mathrm{S}(n) = \mathrm{-M(A,Z)}+\mathrm{M(A-1,Z)}+
      \mathrm{M(n)}\f$: Neutron separation energy \n
      \f$\mathrm{S}(p) = \mathrm{-M(A,Z)}+\mathrm{M(A-1,Z-1)}+
      \mathrm{H^{1}}\f$: Proton separation energy \n
      \f$\mathrm{S}(2n) = \mathrm{-M(A,Z)}+\mathrm{M(A-2,Z)}+
      \mathrm{2M(n)}\f$: Two neutron separation energy \n
      \f$\mathrm{S}(2p) = \mathrm{-M(A,Z)}+\mathrm{M(A-2,Z-2)}+
      \mathrm{2 M(H^{1)}}\f$: Two proton separation energy \n

      \todo The presence or absence of the electron binding energy
      contribution is not so well documented here. 

  */
  class nuclear_mass {
  public:
    
    nuclear_mass();
    
    virtual ~nuclear_mass() {};

    /** 
	\brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N) {
      return true;
    }
    
    /** \brief Fill \c n with the information from nucleus with the given
	neutron and proton number
	
	All masses are given in \f$\mathrm{fm}^{-1}\f$. The total mass
	(withouth the electrons) is put in part::m, and the degeneracy
	(part::g) is arbitrarily set to 1 for even A nuclei and 2 for
	odd A nuclei.
    */
    int get_nucleus(int Z, int N, nucleus &n) {
      n.Z=Z;
      n.N=N;
      n.A=Z+N;
      n.mex=mass_excess(Z,N)/o2scl_const::hc_mev_fm;
      n.m=n.mex+((Z+N)*o2scl_fm::mass_amu-Z*o2scl_fm::mass_electron);
      n.ms=n.mex+((Z+N)*o2scl_fm::mass_amu-Z*o2scl_fm::mass_electron);
      n.be=(n.m-N*o2scl_fm::mass_neutron-Z*o2scl_fm::mass_proton);
      if (n.A%2==0) n.g=1.0;
      else n.g=2.0;
      return 0;
    }
    
    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N) {
      if (form_type==cont_type) {
	return mass_excess_d(Z,N);
      }
      set_err("Using base class version of mass_excess().",gsl_nobase);
      return 0.0;
    }
    

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N) {
      if (form_type==disc_type) {
	int Z1=(int)Z;
	int N1=(int)N;
	double mz1n1=mass_excess(Z1,N1);
	double mz1n2=mass_excess(Z1,N1+1);
	double mz2n1=mass_excess(Z1+1,N1);
	double mz2n2=mass_excess(Z1+1,N1+1);
	double mz1=mz1n1+(N-N1)*(mz1n2-mz1n1);
	double mz2=mz2n1+(N-N1)*(mz2n2-mz2n1);
	return mz1+(Z-Z1)*(mz2-mz1);
      }
      set_err("Using base class version of mass_excess_d().",gsl_nobase);
      return 0.0;
    }
    
    /** 
	\brief Return the binding energy in MeV
	
	The binding energy is defined to be negative for bound 
	nuclei, thus the binding energy per baryon of Pb-208
	is about -8*208 = -1664 MeV.
    */
    virtual double binding_energy(int Z, int N) {
      return (mass_excess(Z,N)+
	      ((Z+N)*o2scl_fm::mass_amu-Z*o2scl_fm::mass_electron-
	       N*o2scl_fm::mass_neutron-Z*o2scl_fm::mass_proton)*
	      o2scl_const::hc_mev_fm);
    }

    /** 
	\brief Return the binding energy in MeV
	
	The binding energy is defined to be negative for bound 
	nuclei, thus the binding energy per baryon of Pb-208
	is about -8*208 = -1664 MeV.
    */
    virtual double binding_energy_d(double Z, double N) {
      return (mass_excess_d(Z,N)+
	      ((Z+N)*o2scl_fm::mass_amu-Z*o2scl_fm::mass_electron-
	       N*o2scl_fm::mass_neutron-Z*o2scl_fm::mass_proton)*
	      o2scl_const::hc_mev_fm);
    }
    
    /** \brief Return the total mass of the nucleus (without the electrons) 
	in MeV
    */
    virtual double total_mass(int Z, int N) {
      return (mass_excess(Z,N)+((Z+N)*o2scl_fm::mass_amu-
				Z*o2scl_fm::mass_electron)*
	      o2scl_const::hc_mev_fm);
    }

    /** \brief Return the total mass of the nucleus (without the electrons) 
	in MeV
    */
    virtual double total_mass_d(double Z, double N) {
      return (mass_excess_d(Z,N)+((Z+N)*o2scl_fm::mass_amu-
				  Z*o2scl_fm::mass_electron)*
	      o2scl_const::hc_mev_fm);
    }

    /// Return Z given the element name
    int eltoZ(std::string el) {
      return element_table.find(el)->second;
    }
    
    /// Return the element name given Z
    std::string Ztoel(int Z) {
      return element_list[Z];
    }
    
    /** 
	\brief Parse a string of the form "Pb208"
	
	Note that this does not correctly interpret dashes,
	e.g. "Pb-208".  will not work.
     */
    int parse_elstring(std::string ela, int &Z, int &N, int &A) {
      if (ela.length()<2) {
	set_err_ret
	  ("Element name too short in nuclear_mass::parse_elstring().",
	   gsl_efailed);
      }
      if (ela.length()>3 && isalpha(ela[2])) {
	std::string el=ela.substr(0,3);
	Z=eltoZ(el);
	A=stoi(ela.substr(3,ela.length()-2));
	N=A-Z;
	return 0;
      }
      if (ela.length()>2 && isalpha(ela[1])) {
	std::string el=ela.substr(0,2);
	Z=eltoZ(el);
	A=stoi(ela.substr(2,ela.length()-2));
	N=A-Z;
	return 0;
      }
      std::string el=ela.substr(0,1);
      Z=eltoZ(el);
      A=stoi(ela.substr(1,ela.length()-1));
      N=A-Z;
      return 0;
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// String comparison operator for element_table
    struct string_less_than {
      bool operator()(const std::string s1, const std::string s2) const {
	return s1<s2;
      }
    };
    
    /// The number of elements (proton number)
    static const int nelements=119;
    
    /// A map containing the proton numbers organized by element name
    std::map<std::string,int,string_less_than> element_table;
    
    /// A convenient typedef for an iterator for element_table
    typedef std::map<std::string,int,string_less_than>::iterator table_it;
    
    /// The list of elements organized by proton number
    std::string element_list[nelements];

    /** 
	\name Indicate whether or not the mass formula is 
	discrete or continuous
    */
    //@{
    int form_type;
    static const int cont_type=1;
    static const int disc_type=2;
    //@}

#endif
    
  };

  /** 
      \brief Fittable mass formula
  */
  class nuclear_mass_fit : public nuclear_mass {
    
  public:
    
    /// Number of fitting parameters
    size_t nfit;

    /// Fix parameters from an array for fitting
    virtual int fit_fun(size_t nv, const ovector_view &x) {
      return 0;
    }
    
    /// Fill array with guess from present values for fitting
    virtual int guess_fun(size_t nv, ovector_view &x) {
      return 0;
    }

  };

  /** 
      \brief Semi-empirical mass formula

      A naive semi-empirical mass formula. 

      \note The default parameters are arbitrary, and are not
      determined from a fit.
  */
  class semi_empirical_mass : public nuclear_mass_fit {

  public:

    /// Binding energy (negative and in MeV, default -16)
    double B;

    /// Symmetry energy (in MeV, default 23.7)
    double Sv;

    /// Surface energy (in MeV, default 18)
    double Ss;

    /// Coulomb energy (in MeV, default 0.7)
    double Ec;

    /// Pairing energy (MeV, default 13.0)
    double Epair;
    
    semi_empirical_mass() {
      B=-16.0;
      Ss=18.0;
      Ec=0.7;
      Sv=23.7;
      Epair=13.0;
      nfit=5;
      form_type=cont_type;
    }

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess_d(double Z, double N) {
      double A=Z+N, cA=cbrt(A);
      double EoA=B+Ss/cA+Ec*Z*Z/cA/A+Sv*pow(1.0-2.0*Z/A,2.0);
      if (((int)(A-Z+0.01))%2==0 && ((int)(Z+0.01))%2==0) {
	EoA-=Epair/pow(A,1.5);
      } else if (((int)(A-Z+0.01))%2==1 && ((int)(Z+0.01))%2==1) {
	EoA+=Epair/pow(A,1.5);
      }
      double ret=EoA*A+Z*(o2scl_fm::mass_proton+o2scl_fm::mass_electron)*
	o2scl_const::hc_mev_fm+N*o2scl_fm::mass_neutron*
	o2scl_const::hc_mev_fm-A*o2scl_fm::mass_amu*o2scl_const::hc_mev_fm;
      return ret;
    }
    
    /// Fix parameters from an array for fitting
    virtual int fit_fun(size_t nv, const ovector_view &x) {
      B=-x[0]; Sv=x[1]; Ss=x[2]; Ec=x[3]; Epair=x[4];
      return 0;
    }

    /// Fill array with guess from present values for fitting
    virtual int guess_fun(size_t nv, ovector_view &x) {
      x[0]=-B; x[1]=Sv; x[2]=Ss; x[3]=Ec; x[4]=Epair;
      return 0;
    }
    
  };
  
  /** 
      \brief Mass formula entry structure for Moller, et al.
  */
  typedef struct mnms95_mass_entry_s {
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /** \name Ground state deformations (perturbed-spheroid parameterization)
     */
    //@{
    /// Quadrupole
    double eps2;
    /// Octupole
    double eps3;
    /// Hexadecapole
    double eps4;
    /// Hexacontatetrapole
    double eps6;
    /// Hexacontatetrapole without mass asymmetry
    double eps6sym;
    //@}
    
    /** \name Ground state deformations in the spherical-harmonics expansion
     */
    //@{
    /// Quadrupole
    double beta2;
    /// Octupole
    double beta3;
    /// Hexadecapole
    double beta4;
    /// Hexacontatetrapole
    double beta6;
    //@}
    
    /// The ground-state microscopic energy
    double Emic;
    
    /// The theoretical mass excess (in MeV)
    double Mth;
    
    /// The experimental mass excess (in MeV)
    double Mexp;
    
    /// Experimental mass excess error
    double sigmaexp;
    
    /// The ground-state microscopic energy in the FRLDM
    double EmicFL;
    
    /// The theoretical mass excess in the FRLDM
    double MthFL;
    
  } mnms95_mass_entry;
  
  template<> int io_tlate<mnms95_mass_entry>::input
    (cinput *co, in_file_format *ins, mnms95_mass_entry *t);
  template<> int io_tlate<mnms95_mass_entry>::output
    (coutput *co, out_file_format *outs, mnms95_mass_entry *t);
  template<> const char *io_tlate<mnms95_mass_entry>::type();
  
  typedef io_tlate<mnms95_mass_entry> mnms95_mass_entry_io_type;
  
  /** 
      \brief Mass formula from Moller, Nix, Myers and Swiatecki
      
      The data containing an object of type moller_mass_entry for 8979
      nuclei is automatically loaded by the constructor.  If the file
      (nndc/moller.dat.gz) is not found, then is_loaded() will return
      false and all calls to get_ZN() will return an object with \c
      N=Z=0.

      \todo Some the blank entries in the table have been replaced
      with zero. This is confusing, so it needs to be fixed by
      regenerating the data file and somehow correctly representing
      the blank entries.
      
  */
  class mnms95_mass : public nuclear_mass {
  public:
    
    mnms95_mass();

    /** 
	\brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N) {
      mnms95_mass_entry ret;
      ret=get_ZN(Z,N);
      if (ret.Z==0 && ret.N==0) return 0.0;
      return ret.Mth;
    }
    
    /** 
	\brief Get the entry for the specified proton and neutron number
	
	This method searches the table using a cached binary search
	algorithm. It is assumed that the table is sorted first by
	proton number and then by neutron number.
    */
    mnms95_mass_entry get_ZN(int l_Z, int l_N);
    
    /// Verify that the constructor properly loaded the table
    bool is_loaded() { return loaded; };
    
    /// The number of table entries
    int n;
    
    /// The array containing the table
    mnms95_mass_entry *mass;

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// True if the table was successfully loaded
    bool loaded;
    
    /// The last table index for caching
    int last;
    
#endif
    
  };
  
  /// The experimental values from Moller, Nix, Myers and Swiatecki
  class mnms95_mass_exp : public mnms95_mass {

  public:

    /** 
	\brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N) {
      mnms95_mass_entry ret;
      ret=get_ZN(Z,N);
      if (ret.Z==0 && ret.N==0) return 0.0;
      return ret.Mexp;
    }

  };

  /** 
      \brief Atomic mass entry structure
  */
  typedef struct ame_entry_s {
    
    /// N-Z
    int NMZ;
    
    /// Neutron number
    int N;
    
    /// Proton number
    int Z;
    
    /// Atomic number
    int A;
    
    /// Element name
    std::string el;
    
    /// Data origin
    std::string orig;
    
    /// Mass excess
    double mass;
    
    /// Mass excess uncertainty
    double dmass;
    
    /// Binding energy (given in the '95 data)
    double be;
    
    /// Binding energy uncertainty (given in the '95 data)
    double dbe;
    
    /// Binding energy / A (given in the '03 data)
    double beoa;
    
    /// Binding energy / A uncertainty (given in the '03 data)
    double dbeoa;
    
    /// Beta decay mode
    std::string bdmode;
    
    /// Beta-decay energy
    double bde;

    /// Beta-decay energy uncertainty
    double dbde;
    
    /// ?
    int A2;
    
    /// Atomic mass
    double amass;
    
    /// Atomic mass uncertainty
    double damass;
    
  } ame_entry;
  
  /** \brief A support class for I/O of the 1995 AME data
   */
  class ame_entry95_io_type : public io_tlate<ame_entry> {
  public:
    int input(cinput *co, in_file_format *ins, ame_entry *t);
    int output(coutput *co, out_file_format *outs, ame_entry *t);
    virtual const char *type();
  };
  
  /** \brief A support class for I/O of the 2003 AME data
   */
  class ame_entry03_io_type : public io_tlate<ame_entry> {
  public:
    int input(cinput *co, in_file_format *ins, ame_entry *t);
    int output(coutput *co, out_file_format *outs, ame_entry *t);
    virtual const char *type();
  };
  
  /** 
      \brief Mass formula from the Atomic Mass Evaluation (2005 and 1993)
      
      This class provides an interface to the atomic mass table using
      data from \ref Audi95 and \ref Audi03. The data is stored
      in ame::mass, which is an array of type ame_entry.
      
      There are four data sets, selected by the specification of the
      \c version string in the constructor.
      - "95rmd" - "Recommended" data from \ref Audi95 (ame95rmd.dat.gz)
      - "95exp" - "Experimental" data from \ref Audi95 (ame95exp.dat.gz)
      - "03round" - "Rounded" data from \ref Audi03 (ame03round.dat.gz)
      - "03" - Data from \ref Audi03 (default) (ame03.dat.gz)
      
      If any string other than these four is used, the default
      data is loaded. If the constructor cannot find the data file
      (e.g. because of a broken installation), then ame::is_loaded()
      returns false.
      
      The 1995 data provided the binding energy stored in
      ame_entry::be and ame_entry::dbe, while the 2003 data provided
      the binding energy divided by the atomic number stored in
      ame_entry::beoa and ame_entry::dbeoa. When the 1995 data is used
      ame_entry::beoa and ame_entry::dbeoa are calculated manually,
      and when the 2003 data is used ame_entry::be and ame_entry::dbe
      are calculated manually.
      
      Note that blank entries in the original table that correspond to
      columns represented by the type \c double are set to zero
      arbitrarily.
      
      Note that all uncertainties are 1 sigma uncertainties.
      
      \future Create a caching and more intelligent search system for
      the table. The table is sorted by A and then N, so we could
      probably just copy the search routine from mnms95_mass, which is
      sorted by Z and then N.
      \future There are strict definitions of the atomic mass unit
      and other constants that are defined by the 1995 and 2003
      atomic mass evaluations. These should be included properly.
  */
  class ame_mass : public nuclear_mass {
  public:

    /// Create a collection specified by \c version
    ame_mass(std::string version="");
    
    /** 
	\brief Return false if the mass formula does not include 
	specified nucleus
    */
    virtual bool is_included(int Z, int N);

    /// Given \c Z and \c N, return the mass excess in MeV
    virtual double mass_excess(int Z, int N) {
      ame_entry ret;
      ret=get_ZN(Z,N);
      if (ret.Z==0 && ret.N==0) return 0.0;
      return ret.mass/1.0e3;
    }
    
    /// Get element with Z=l_Z and N=l_N (e.g. 82,126).
    ame_entry get_ZN(int l_Z, int l_N);
    
    /// Get element with Z=l_Z and A=l_A (e.g. 82,208).
    ame_entry get_ZA(int l_Z, int l_A);
    
    /// Get element with name l_el and A=l_A (e.g. "Pb",208).
    ame_entry get_elA(std::string l_el, int l_A);
    
    /// Get element with string (e.g. "Pb208")
    ame_entry get(std::string nucleus);
    
    /// The number of entries (about 3000).
    int n;
    
    /// The short names of the columns (length 16)
    std::string *short_names;
    
    /// The long names of the columns (length 16)
    std::string *col_names;
    
    /// The reference for the original data
    std::string reference;
    
    /// The array containing the mass data of length ame::n
    ame_entry *mass;
    
    /// Returns true if the constructor succesfully loaded the data
    bool is_loaded() { return loaded; }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// True if loading the data was successful
    bool loaded;

#endif
    
  };
  
#ifndef DOXYGENP
}
#endif

#endif
