/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_GSL_ASTEP_H
#define O2SCL_GSL_ASTEP_H

#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv.h>

#include <o2scl/adapt_step.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Adaptive stepper (GSL)
      
      \todo Finish implementing the scaled "control"
      \todo Allow user to find out how many steps were taken, etc.
  */
  template<class param_t, class func_t, class vec_t=ovector_view,
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc> class gsl_astep : 
  public adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> {
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// Size of allocated vectors
    size_t ndim;
    
    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;

    /// Number of scalings
    size_t sdim;
    
    /// Scalings
    double *scale_abs;

    /** 
	\brief The GSL evolution object for \ref gsl_astep 
	[protected subclass]
    */
    class gsl_odeiv_evolve {
    public:
	
      /// Number of differential equations
      size_t dimension;

      /// The values of the functions
      alloc_vec_t y0;
      /// The uncertainty
      alloc_vec_t yerr;
      /// The input derivatives
      alloc_vec_t dydt_in;
      /// The output derivatives
      alloc_vec_t dydt_out;
      /// Size of the memory allocation
      size_t ndim;
      
      /// The size of the last step
      double last_step;
      /// The number of steps (?)
      unsigned long int count;
      /// The number of failed steps
      unsigned long int failed_steps;
    };

    /// Automatically adjust step-size
    int hadjust(size_t dim, unsigned int ord, const vec_t &y,
		vec_t &yerr, vec_t &yp, double *h) 
    {
	
      gsl_ode_control *state = &con;
	
      const double eps_abs = state->eps_abs;
      const double eps_rel = state->eps_rel;
      const double a_y     = state->a_y;
      const double a_dydt  = state->a_dydt;
	
      const double S = 0.9;
      const double h_old = *h;
      
      double rmax = DBL_MIN;
      size_t i;
	
      for(i=0; i<dim; i++) {
	double D0;
	if (state->standard || sdim==0) {
	  D0=eps_rel*(a_y*fabs(y[i])+a_dydt*fabs(h_old*yp[i]))+eps_abs;
	} else {
	  D0=eps_rel*(a_y*fabs(y[i])+a_dydt*fabs(h_old*yp[i]))+
	    eps_abs*scale_abs[i%sdim];
	}
	const double r=fabs(yerr[i]) / fabs(D0);
	rmax = GSL_MAX_DBL(r, rmax);
      }
	
      if(rmax > 1.1) {

	/* decrease step, no more than factor of 5, but a fraction S more
	   than scaling suggests (for better accuracy) */

	double r =  S / pow(rmax, 1.0/ord);
	  
	if (r < 0.2) {
	  r = 0.2;
	}
	  
	*h = r * h_old;
	  
	return GSL_ODEIV_HADJ_DEC;

      } else if (rmax < 0.5) {

	/* increase step, no more than factor of 5 */
	double r = S / pow(rmax, 1.0/(ord+1.0));
	  
	if (r > 5.0) {
	  r = 5.0;
	}
	  
	/* don't allow any decrease caused by S<1 */
	if (r < 1.0) {
	  r = 1.0;
	}
	  
	*h = r * h_old;

	return GSL_ODEIV_HADJ_INC;

      } else {

	/* no change */
	return GSL_ODEIV_HADJ_NIL;

      }

    }

    /// Apply the evolution for the next adaptive step
    int evolve_apply(gsl_odeiv_evolve *e, size_t nvar, param_t &pa, 
		     func_t &derivs, double *t, double t1, double *h, 
		     vec_t &y) 
    {
      double t0 = *t;
      double h0 = *h;
      int step_status;
      int final_step = 0;
      /* remaining time, possibly less than h */
      double dt = t1 - t0;  
	
      if ((dt < 0.0 && h0 > 0.0) || (dt > 0.0 && h0 < 0.0)) {
	GSL_ERROR ("step direction must match interval direction", 
		   GSL_EINVAL);
      }
	
      for(size_t ic=0;ic<nvar;ic++) e->y0[ic]=y[ic];
	
      /* Calculate initial dydt  */
      int status=derivs(t0,nvar,y,e->dydt_in,pa);
      if (status) return status;
	
      try_step:
  
      if ((dt >= 0.0 && h0 > dt) || (dt < 0.0 && h0 < dt)) {
	h0 = dt;
	final_step = 1;
      } else{
	final_step = 0;
      }
      
      step_status=this->stepp->step(t0,h0,nvar,y,e->dydt_in,y,e->yerr,
				    e->dydt_out,pa,derivs);
	
      /* Check for stepper internal failure */
      if (step_status != gsl_success)	{
	// Notify user which step-size caused the failure
	*h=h0;
	return step_status;
      }
  
      e->count++;
      e->last_step = h0;
      if (final_step) {
	*t = t1;
      } else{
	*t = t0 + h0;
      }
  
      /* Check error and attempt to adjust the step. */
      const int hadjust_status=hadjust(nvar,this->stepp->get_order(),
				       y,e->yerr,e->dydt_out,&h0);
	
      if (hadjust_status == GSL_ODEIV_HADJ_DEC) {
	  
	/* Step was decreased. Undo and go back to try again. */
	for(size_t ic=0;ic<nvar;ic++) y[ic]=e->y0[ic];
	e->failed_steps++;
	goto try_step;
      }
	
      /* suggest step size for next time-step */
      *h = h0;  
  
      return step_status;
    }
      
#endif
    
    public:
      
    /// Control structure
    typedef struct
    {
      /// Absolute precision
      double eps_abs;
      /// Relative precision
      double eps_rel;
      /// Function scaling factor
      double a_y;
      /// Derivative scaling factor
      double a_dydt;
      /// Use standard or scaled algorithm
      bool standard;

    } gsl_ode_control;

    /// Control specification
    gsl_ode_control con;
    
    gsl_astep() {

      this->verbose=0;

      con.eps_abs=1.0e-6;
      con.eps_rel=0.0;
      con.a_y=1.0;
      con.a_dydt=0.0;
      con.standard=true;

      ndim=0;
      sdim=0;
    }
      
    virtual ~gsl_astep() {
      if (sdim!=0) delete[] scale_abs;
    }
    
    /// Set scalings
    template<class svec_t> int set_scale(size_t nscal, const svec_t &scale) {
      if (nscal>=sdim) {
	if (sdim!=0) delete[] scale_abs;
	scale_abs=new double[nscal];
	sdim=nscal;
      }
      for(size_t i=0;i<nscal;i++) {
	scale_abs[i]=scale[i];
      }
      return 0;
    }

    /** \brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c x
	of an \c n-dimensional system \c derivs starting with \c y. On
	exit, \c x and \c y contain the new values at the end of the
	step and \c h contains the size of the step. 
    */
    virtual int astep(double &x, double &h, double xmax, size_t n, 
		      vec_t &y, param_t &pa, func_t &derivs) 
    {

      // The evolution object
      gsl_odeiv_evolve goe;
      
      ao.allocate(goe.y0,n);
      ao.allocate(goe.yerr,n);
      ao.allocate(goe.dydt_in,n);
      ao.allocate(goe.dydt_out,n);
	
      // Set up odeiv_evolve object
      goe.dimension=n;

      goe.count=0;
      goe.failed_steps=0;
      goe.last_step=0.0;
      
      int ret=evolve_apply(&goe,n,pa,derivs,&x,xmax,&h,y);
	
      if (this->verbose>0) {
	std::cout << x << " ";
	for(size_t j=0;j<n;j++) std::cout << y[j] << " ";
	std::cout << std::endl;
      }

      ao.free(goe.y0);
      ao.free(goe.yerr);
      ao.free(goe.dydt_in);
      ao.free(goe.dydt_out);
	
      return ret;
    }

    /** \brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c x
	of an \c n-dimensional system \c derivs starting with \c y. On
	exit, \c x and \c y contain the new values at the end of the
	step and \c h contains the size of the step. 
    */
    virtual int astep_derivs(double &x, double &h, double xmax, size_t n, 
			     vec_t &y, vec_t &dydx, param_t &pa, 
			     func_t &derivs) 
    {

      // The evolution object
      gsl_odeiv_evolve goe;
      
      ao.allocate(goe.y0,n);
      ao.allocate(goe.yerr,n);
      ao.allocate(goe.dydt_in,n);
      ao.allocate(goe.dydt_out,n);
	
      // Set up odeiv_evolve object
      goe.dimension=n;

      goe.count=0;
      goe.failed_steps=0;
      goe.last_step=0.0;
      
      int ret=evolve_apply(&goe,n,pa,derivs,&x,xmax,&h,y);

      // Compute the final derivatives
      derivs(x,n,y,dydx,pa);
	
      if (this->verbose>0) {
	std::cout << x << " ";
	for(size_t j=0;j<n;j++) std::cout << y[j] << " ";
	std::cout << std::endl;
      }

      ao.free(goe.y0);
      ao.free(goe.yerr);
      ao.free(goe.dydt_in);
      ao.free(goe.dydt_out);
	
      return ret;
    }
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif
