/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IV_SOLVE_H
#define O2SCL_ODE_IV_SOLVE_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/adapt_step.h>
#include <o2scl/gsl_astep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Base for solving initial-value ODE problems
      
      \todo Consider modifying so that this can handle tables
      which are two small by removing o2sclf the rows and doubling
      the stepsize.
      \todo Decide what to do if the adaptive stepper fails. (1/18/07 
      - two approaches: continue no matter what, or just stop)
      \todo Convert to using \c astep_derivs()?
      
  */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc, 
    class adapt_step=gsl_astep<param_t,func_t,vec_t,alloc_vec_t,alloc_t> > 
    class ode_iv_solve 
    
    {
      public:
      
      ode_iv_solve() {
	verbose=0;
      }
      
      virtual ~ode_iv_solve() {}
      
      /** 
	  \brief Solve the initial-value problem and output a table 

	  Initially, \c xsol should be a vector of size \c nsol, and
	  \c ysol should be a two-dimensional array
	  (i.e. omatrix_view) of size \c [nsol][n]. On exit, \c nsol
	  will will be the size of the solution table, less than or
	  equal to the original value of \c nsol.
      */
      template<class mat_t>
      int solve_table(double x0, double x1, double h, size_t n,
		      vec_t &ystart, size_t &nsol, vec_t &xsol, 
		      mat_t &ysol, param_t &pa, func_t &derivs) 
      {
	int ret=0, nmax=nsol, i;
	size_t j;
  
	xsol[0]=x0;
	for(j=0;j<n;j++) ysol[0][j]=ystart[j];
	if (verbose>0) print_iter(xsol[0],n,ystart);
  
	for(i=1;i<nmax && xsol[i-1]<x1 && ret==0;i++) {
    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
    
	  omatrix_row ar(ysol,i);
	  ret=astepper.astep(xsol[i],h,x1,n,ar,pa,derivs);
    
	  if (verbose>0) print_iter(xsol[i],n,ar);
	}

	nsol=i;

	if (ret!=0) return ret;

	if (i==nmax && xsol[i-1]<x1) {
	  set_err("Ran out of space in ode_iv_solve::solve_table().",
		  gsl_etable);
	  return gsl_etable;
	}
	return 0;
      }

      /** \brief Solve the initial-value problem from \c x0 
	  to \c x1 over a grid 

	  Initially, \c xsol should be an array of size \c nsol, and \c
	  ysol should be a \c omatrix of size \c [nsol][n].

	  This function never takes a step larger than the grid
	  size. This could cause inaccuracy if the grid size is too
	  fine.
      */
      template<class mat_t>
      int solve_grid(double x0, double x1, double h, size_t n,
		     vec_t &ystart, size_t nsol, vec_t &xsol, mat_t &ysol,
		     param_t &pa, func_t &derivs) 
      {
	double x=x0, xnext;
	int ret=0;
	size_t j;

	xsol[0]=x0;
	for(j=0;j<n;j++) ysol[0][j]=ystart[j];
	if (verbose>0) print_iter(xsol[0],n,ystart);

	for(size_t i=1;i<nsol && ret==0;i++) {
    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];

	  xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
    
	  while(xsol[i]<xnext && ret==0) {
	    omatrix_row ar(ysol,i);
	    ret=astepper.astep(xsol[i],h,xnext,n,ar,pa,derivs);
	    if (verbose>0) print_iter(xsol[i],n,ar);
	  }

	}
	if (ret!=0) return ret;
	return 0;
      }

      /** \brief Solve the initial-value problem 
	  to get the final value 

	  For a partiular adaptive stepper, this will likely be the
	  fastest approach, but it provides no information about 
	  the solution other than the final value at x=x1.
      */
      int solve_final_value(double x0, double x1, double h, size_t n,
			    vec_t &ystart, vec_t &yend, 
			    param_t &pa, func_t &derivs) 
      {
	double x=x0;
	int ret=0;
	if (verbose>0) print_iter(x0,n,ystart);

	for(size_t i=0;i<n;i++) yend[i]=ystart[i];

	while(x<x1 && ret==0) {
	  
	  ret=astepper.astep(x,h,x1,n,yend,pa,derivs);
	  
	  if (verbose>0) print_iter(x,n,yend);
	}

	if (ret!=0) return ret;
	return 0;
      }
      
      /// Set output level
      int verbose;

      /// The adaptive stepper utilized
      adapt_step astepper;
      
#ifndef DOXYGEN_INTERNAL

      protected:
      
      /// Print out iteration information
      virtual int print_iter(double x, size_t nv, vec_t &y) {
	std::cout << x << " ";
	for(size_t i=0;i<nv;i++) std::cout << y[i] << " ";
	std::cout << std::endl;
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
	return 0;
      }
  
#endif

    };

#ifndef DOXYGENP
}
#endif

#endif
