/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_LANCZOS_H
#define O2SCL_LANCZOS_H

#include <cmath>
#include <gsl/gsl_linalg.h>
#include <o2scl/collection.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Lanczos diagonalization
      
      This is useful for approximating the largest eigenvalues of a
      symmetric matrix.

      The vector and matrix types can be any type which provides
      access via \c operator[], given suitably constructed allocation
      types.
      
      The tridiagonalization routine was rewritten from the EISPACK
      routines \c imtql1.f (but uses \c gsl_hypot() instead of \c
      pythag.f). 

      \future The function eigen_tdiag() automatically sorts the
      eigenvalues, which may not be necessary.
  */
  template<class vec_t, class mat_t, class alloc_vec_t, class alloc_t>
    class lanczos {
  public:

    lanczos() {
      td_iter=30;
      td_lasteval=0;
    }
  
    /** \brief Number of iterations for finding the eigenvalues of the
	tridiagonal matrix (default 30)
    */
    size_t td_iter;

    /** \brief The index for the last eigenvalue not determined if
	tridiagonalization fails
    */
    size_t td_lasteval;

    /** \brief Approximate the largest eigenvalues of a symmetric
	matrix \c mat using the Lanczos method

	Given a square matrix \c mat with size \c size, this function
	applies \c n_iter iterations of the Lanczos algorithm to
	produce \c n_iter approximate eigenvalues stored in \c
	eigen. As a by-product, this function also partially
	tridiagonalizes the matrix placing the result in \c diag and
	\c off_diag. Before calling this function, space must have
	already been allocated for \c eigen, \c diag, and \c
	off_diag. All three of these arrays must have at least enough
	space for \c n_iter elements.
	
	Choosing /c n_iter = \c size will produce all of the exact
	eigenvalues and the corresponding tridiagonal matrix, but this
	may be slower than diagonalizing the matrix directly.
    */
    int eigenvalues(size_t size, mat_t &mat, size_t n_iter, 
		    vec_t &eigen, vec_t &diag, vec_t &off_diag) {
      double t;
      bool cont=true;
      size_t i, j, k;

      alloc_vec_t v;
      alloc_vec_t w;
      alloc_vec_t b3;
      alloc_vec_t prod;
      alloc_t ao;
      ao.allocate(v,size);
      ao.allocate(w,size);
      ao.allocate(b3,size);
      ao.allocate(prod,size);

      // Pick a unit vector
      w[0]=1.0;
      for(i=1;i<size;i++) w[i]=0.0;

      for(i=0;i<size;i++) v[i]=0;
      j=0;
      while (cont) {
	if (j!=0) {
	  for(i=0;i<size;i++) {
	    t=w[i];
	    w[i]=v[i]/off_diag[j-1];
	    v[i]=-off_diag[j-1]*t;
	  }
	}
	product(size,mat,w,prod);
	for(k=0;k<size;k++) v[k]+=prod[k];
	diag[j]=0.0;
	off_diag[j]=0.0;

	for(k=0;k<size;k++) diag[j]+=w[k]*v[k];
	for(k=0;k<size;k++) v[k]-=diag[j]*w[k];
	for(k=0;k<size;k++) off_diag[j]+=v[k]*v[k];
	off_diag[j]=sqrt(off_diag[j]);
	j++;
    
	if (j>=n_iter || off_diag[j-1]==0.0) cont=false;
    
	if (j>0) {
	  for(k=0;k<size-1;k++) {
	    b3[k+1]=off_diag[k];
	    eigen[k]=diag[k];
	  }
	  eigen[size-1]=diag[size-1];
	  if (eigen_tdiag(j,eigen,b3)!=0) {

	    ao.free(v);
	    ao.free(w);
	    ao.free(b3);
	    ao.free(prod);

	    add_err_ret("Call to eigen_tdiag() in eigenvalues() failed.",
			gsl_efailed);
	  }
	}
      }

      ao.free(v);
      ao.free(w);
      ao.free(b3);
      ao.free(prod);
  
      return 0;
    }
    
    /** 
	\brief In-place diagonalization of a tri-diagonal matrix

	On input, the vectors \c diag and \c off_diag should both be
	vectors of size \c n. The diagonal entries stored in \c diag,
	and the \f$ n-1 \f$ off-diagonal entries should be stored in
	\c off_diag, starting with \c off_diag[1].  The value in \c
	off_diag[0] is unused. The vector \c off_diag is destroyed by
	the computation.

	This uses an implict QL method from the EISPACK routine \c
	imtql1. The value of \c ierr from the original Fortran routine
	is stored in \ref td_lasteval.

     */
    int eigen_tdiag(size_t n, vec_t &diag, vec_t &off_diag) {
      int i,j,l,m,mml;
      double b,c,f,g,p,r,s,tst1,tst2;

      if (n==1) return 0;
      
      for(size_t ij=1;ij<n;ij++) off_diag[ij-1]=off_diag[ij];
      off_diag[n-1]=0.0;
  
      bool done=false;
  
      l=1;
      j=0;
  
      while (done==false && l<=((int)n)) {
    
	// Look for small sub-diagonal element
	bool idone=false;
	for(m=l;m<((int)n) && idone==false;m++) {
	  tst1=fabs(diag[m-1])+fabs(diag[m]);
	  tst2=tst1+fabs(off_diag[m-1]);
	  if (tst2==tst1) {
	    m--;
	    idone=true;
	  }
	}
    
	p=diag[l-1];

	if (m!=l && j==((int)td_iter)) {
	  
	  // Set error. No convergence after td_iter iterations
	  td_lasteval=l;
	  set_err_ret("No convergence in lanczos::eigen_tdiag()",
		      gsl_efailed);
	}

	if (m!=l) {

	  j++;

	  // Form shift
	  g=(diag[l]-p)/(2.0*off_diag[l-1]);
	  r=gsl_hypot(g,1.0);
      
	  g=diag[m-1]-p+off_diag[l-1]/(g+(g>=0.0 ? fabs(r) : -fabs(r)));
	  s=1.0;
	  c=1.0;
	  p=0.0;
	  mml=m-l;
      
	  for(int ii=1;ii<=mml;ii++) {

	    i=m-ii;
	    f=s*off_diag[i-1];
	    b=c*off_diag[i-1];
	    r=gsl_hypot(f,g);
	    off_diag[i]=r;
	
	    if (r==0.0) {

	      // Recover from underflow
	      diag[i]-=p;
	      off_diag[m-1]=0.0;
	      ii=mml+1;

	    } else {

	      s=f/r;
	      c=g/r;
	      g=diag[i]-p;
	      r=(diag[i-1]-g)*s+2.0*c*b;
	      p=s*r;
	      diag[i]=g+p;
	      g=c*r-b;

	    }

	  }
      
	  diag[l-1]-=p;
	  off_diag[l-1]=g;
	  off_diag[m-1]=0.0;
      

	} else {

	  // Order eigenvalues

	  if (l==1) {

	    i=1;
	    diag[i-1]=p;
	
	  } else {

	    bool skip=false;
	    for(int ii=2;ii<=l;ii++) {
	      i=l+2-ii;
	      if (p>=diag[i-2]) {
		ii=l+1;
		skip=true;
	      } else {
		diag[i-1]=diag[i-2];
	      }
	    }
	
	    if (skip==false) i=1;
	    diag[i-1]=p;
	  }

	  j=0;
	  l++;
	}
    
      }
  
      return 0;
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:
      
    /** 
	\brief Naive matrix-vector product
	
	It is assumed that memory is already allocated for \c prod.
    */
    void product(size_t n, mat_t &a, vec_t &w, vec_t &prod) {
      size_t i, j;
      for(i=0;i<n;i++) {
	prod[i]=0.0;
	for(j=0;j<n;j++) {
	  prod[i]+=a[i][j]*w[j];
	}
      }
      return;
    }
    
#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif
