/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TWOD_INTP_H
#define O2SCL_TWOD_INTP_H

#include <iostream>
#include <string>
#include <o2scl/smart_interp.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Two-dimensional interpolation class 

      This class implements
      two-dimensional interpolation. Derivatives and integrals
      along both x- and y-directions can be computed.

      The storage for the data, including the arrays \c x_fun 
      and \c y_fun are all managed by the user. If the data is
      changed without calling reset_interp(), then interp() will
      return incorrect results.

      No caching is done (except for the caching already present
      in sm_interp), and successive calls to interp() with the
      same values of x and y will take nearly the same amount of time.
      
      \todo Test non-square data
      \todo Allow specification of data as [ny][nx] or as [nx][ny]?

      \todo Need to rewrite to use \ref o2scl_interp after that
      class is finished.
      \todo Could also include mixed second/first derivatives:
      deriv_xxy and deriv_xyy.

  */
  class twod_intp {
  public:
    
    twod_intp();

    virtual ~twod_intp();
    
    /** \brief Initialize the data for the 2-dimensional interpolation
	
        The interpolation type (passed directly to int_type) is
	specified in \c int_type and the data is specified in
	\c data. The data should be arranged so that the first
	array index is the y-value (the "row") and the second array
	index is the x-value (the "column"). The arrays \c xfun 
	and \c yfun specify the two independent
	variables. \c xfun should be an array of length
	\c nx , and should be an array of length \c ny. The
	array \c data should be a two-dimensional array of size
	[ny][nx].

	If \c x_first is true, then set_data() creates interpolation
	objects for each of the rows. Calls to interp() then uses
	these to create a column at the specified value of \c x. An
	interpolation object is created at this column to find the
	value of the function at the specified value \c y. If \c
	x_first is false, the opposite strategy is employed. These two
	options will give different results (unless linear
	interpolation is explicity selected). Using \c x_first=false
	will be slightly slower, since the data has to be rearranged
	in the opposite order first. In general, if the data is "more
	accurate" in the x direction than in the y direction, it is
	probably better to choose \c x_first=true.
	
	Valid values of \c int_type are:
	- 0 - default (cubic spline)
	- 1 - linear
	- 2 - cubic spline
	- 3 - periodic cubic spline
	- 4 - akima spline
	- 5 - periodic akima spline
    
    */
    int set_data(int n_x, int n_y, ovector &x_fun,
		 ovector &y_fun, omatrix &u_data, 
		 bool x_first=true);
    
    /** \brief Reset the stored interpolation since the data has changed

        This will return an error if the set_data() has not been called
     */
    int reset_interp();

    /** \brief Perform the 2-d interpolation 
     */
    double interp(double x, double y);

    /** \brief Compute the partial derivative in the x-direction
     */
    double deriv_x(double x, double y);

    /** \brief Compute the partial second derivative in the x-direction
     */
    double deriv2_x(double x, double y);

    /** \brief Compute the integral in the x-direction between x=x0
	and x=x1
     */
    double integ_x(double x0, double x1, double y);

    /** \brief Compute the partial derivative in the y-direction
     */
    double deriv_y(double x, double y);

    /** \brief Compute the partial second derivative in the y-direction
     */
    double deriv2_y(double x, double y);

    /** \brief Compute the integral in the y-direction between y=y0
	and y=y1
     */
    double integ_y(double x, double y0, double y1);

    /** \brief Compute the mixed partial derivative 
	\f$ \frac{\partial^2 f}{\partial x \partial y} \f$
    */
    double deriv_xy(double x, double y);
       

#ifndef DOXYGENP

  protected:

    /// The array of sm_interp pointers 
    sm_interp_vec **si;
    /// An array of omatrix_rows
    omatrix_row **ari;
    /// An array of omatrix_cols
    omatrix_col **aci;

    /// The number of x grid points
    int nx;

    /// The number of y grid points
    int ny;

    /// True if the x interpolation should be done first
    bool xfirst;

    /// True if the data has been specified by the user
    bool data_set;

    /// The definition of the x grid
    ovector *xfun;

    /// The definition of the y grid
    ovector *yfun;

    /// The data
    omatrix *data;

    /// The x-y flipped data if we're doing the y-interpolation first
    omatrix yfirst_data;
    
#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif



