/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_ROOT_BRENT_H
#define O2SCL_GSL_ROOT_BRENT_H

#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>
#include <o2scl/funct.h>
#include <o2scl/root.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief One-dimensional root-finding (GSL)

      This class finds the root of a user-specified function. 
      If \ref test_form is 0, then solve_bkt() stops when the size of
      the bracket is smaller than \ref root::tolx. If \ref test_form
      is 1, then the function stops when the residual is less than
      \ref root::tolf. If test_form is 2, then both tests are applied.

      \todo There is some duplication in the variables \c x_lower, 
      \c x_upper, \c a, and \c b, which could be removed.

      \comment
      Note that I used \c instead of \ref to refer to variables above
      since the variables a and b are protected, and not available if
      compiling the documentation without the internal portion.
      \endcomment
  */
  template<class param_t, class func_t=funct<void *>, class dfunc_t=func_t> 
    class gsl_root_brent : public root<param_t,func_t,dfunc_t> 

    {
      
    public:
      
      gsl_root_brent() {
	test_form=0;
	this->over_bkt=true;
      }
    
      /// Return the type, \c "gsl_root_brent".
      virtual const char *type() { return "gsl_root_brent"; }

      /** 
	  \brief Perform an iteration

	  This function always returns \ref gsl_success.

      */
      int iterate(func_t &f) {
      
	double tol, m;
	
	int ac_equal = 0;
	
	if ((fb < 0 && fc < 0) || (fb > 0 && fc > 0)) {
	  ac_equal = 1;
	  c = a;
	  fc = fa;
	  d = b - a;
	  e = b - a;
	}
  
	if (fabs (fc) < fabs (fb)) {
	  ac_equal = 1;
	  a = b;
	  b = c;
	  c = a;
	  fa = fb;
	  fb = fc;
	  fc = fa;
	}
  
	tol = 0.5 * GSL_DBL_EPSILON * fabs (b);
	m = 0.5 * (c - b);
  
	if (fb == 0) {
	  root = b;
	  x_lower = b;
	  x_upper = b;
    
	  return o2scl::gsl_success;
	}
	if (fabs (m) <= tol) {
	  root = b;
    
	  if (b < c) {
	    x_lower = b;
	    x_upper = c;
	  } else {
	    x_lower = c;
	    x_upper = b;
	  }
    
	  return o2scl::gsl_success;
	}
  
	if (fabs (e) < tol || fabs (fa) <= fabs (fb)) {
	  d = m;            /* use bisection */
	  e = m;
	} else {
	  double p, q, r;   /* use inverse cubic interpolation */
	  double s = fb / fa;
    
	  if (ac_equal) {
	    p = 2 * m * s;
	    q = 1 - s;
	  } else {
	    q = fa / fc;
	    r = fb / fc;
	    p = s * (2 * m * q * (q - r) - (b - a) * (r - 1));
	    q = (q - 1) * (r - 1) * (s - 1);
	  }
      
	  if (p > 0) {
	    q = -q;
	  } else {
	    p = -p;
	  }
	  if (2 * p < GSL_MIN (3 * m * q - fabs (tol * q), fabs (e * q))) {
	    e = d;
	    d = p / q;
	  } else {
	    /* interpolation failed, fall back to bisection */
	  
	    d = m;
	    e = m;
	  }
	}
  
	a = b;
	fa = fb;
  
	if (fabs (d) > tol) {
	  b += d;
	} else {
	  b += (m > 0 ? +tol : -tol);
	}
	
	f(b,fb,*params);
  
	/* Update the best estimate of the root and bounds on each
	   iteration */
  
	root = b;
  
	if ((fb < 0 && fc < 0) || (fb > 0 && fc > 0)) {
	  c = a;
	}
	if (b < c) {
	  x_lower = b;
	  x_upper = c;
	} else {
	  x_lower = c;
	  x_upper = b;
	}
  
	return o2scl::gsl_success;
      }
      
      /// Solve \c func in region \f$ x_1<x<x_2 \f$ returning \f$ x_1 \f$.
      virtual int solve_bkt(double &x1, double x2, param_t &pa, 
			    func_t &f) {
	
	int status, iter=0;
	
	set(f,x1,x2,pa);
  
	if (test_form==0) {
	  /// Test the bracket size

	  status=GSL_CONTINUE;
	  while (status==GSL_CONTINUE && iter<this->ntrial) {
      
	    iter++;
	    iterate(f);
	    status=gsl_root_test_interval(x_lower,x_upper,0.0,this->tolx);
      
	    if (this->verbose>0) {
	      double y;

	      f(root,y,pa);
	    
	      print_iter(root,y,iter,fabs(x_upper-x_lower),this->tolx,
			 "gsl_root_brent");
	    }
	  }
	
	} else if (test_form==1) {
	  /// Test the residual

	  status=GSL_CONTINUE;
	  while (status==GSL_CONTINUE && iter<this->ntrial) {
      
	    iter++;
	    iterate(f);

	    double y;

	    f(root,y,pa);

	    if (fabs(y)<this->tolf) status=o2scl::gsl_success;
      
	    status=gsl_root_test_interval(x_lower,x_upper,0.0,this->tolx);
      
	    if (this->verbose>0) {
	      print_iter(root,y,iter,fabs(x_upper-x_lower),this->tolx,
			 "gsl_root_brent");
	    }
	  }


	} else {
	  /// Test the bracket size and the residual

	  status=GSL_CONTINUE;
	  while (status==GSL_CONTINUE && iter<this->ntrial) {
      
	    iter++;
	    iterate(f);
	    status=gsl_root_test_interval(x_lower,x_upper,0.0,this->tolx);
      
	    if (status==o2scl::gsl_success) {
	      double y;
	      
	      f(root,y,pa);
	      
	      if (fabs(y)>=this->tolf) status=GSL_CONTINUE;
	    }
	    
	    if (this->verbose>0) {
	      double y;
	      
	      f(root,y,pa);
	      
	      print_iter(root,y,iter,fabs(x_upper-x_lower),this->tolx,
			 "gsl_root_brent");
	    }
	  }

	}

	x1=root;
  
	if (status!=o2scl::gsl_success) {
	  int ret=o2scl::err_hnd->get_errno();
	  return ret;
	}
	if (iter>=this->ntrial) {
	  set_err_ret("solve_bkt() exceeded maximum number of iterations.",
		      o2scl::gsl_emaxiter);
	}
  
	return o2scl::gsl_success;
      }

      /// The type of convergence test applied: 0, 1, or 2 (default 0)
      int test_form;
     
      /// Get the most recent value of the root
      double get_root() { return root; }
      
      /// Get the lower limit
      double get_lower() { return x_lower; }
      
      /// Get the upper limit
      double get_upper() { return x_upper; }
    
      /** 
	  \brief Set the information for the solver

	  This function always returns \ref gsl_success.
       */
      int set(func_t &ff, double lower, double upper, param_t &pa) {
    
	if (lower > upper) {
	  set_err_ret("Invalid interval in gsl_root_brent::set().",
		      o2scl::gsl_einval);
	}
      
	params=&pa;
  
	x_lower=lower;
	x_upper=upper;
	root=0.5*(lower+upper);
  
	double f_lower, f_upper;
  
	ff(x_lower,f_lower,pa);
	ff(x_upper,f_upper,pa);
	
	a = x_lower;
	fa = f_lower;
	
	b = x_upper;
	fb = f_upper;
	
	c = x_upper;
	fc = f_upper;
	
	d = x_upper - x_lower;
	e = x_upper - x_lower;
	
	if ((f_lower < 0.0 && f_upper < 0.0) || 
	    (f_lower > 0.0 && f_upper > 0.0)) {
	  set_err_ret
	    ("Endpoints don't straddle y=0 in gsl_root_brent::set().",
	     o2scl::gsl_einval);
	}
	
	return o2scl::gsl_success;
	
      }

#ifndef DOXYGEN_INTERNAL
      
    protected:
      
      /// The present solution estimate
      double root;
      /// The present lower limit
      double x_lower;
      /// The present upper limit
      double x_upper;
      /// The function parameters
      param_t *params;

      /// \name Storage for solver state
      //@{
      double a, b, c, d, e;
      double fa, fb, fc;
      //@}
      
#endif
 
    };
   
#ifndef DOXYGENP
  template<> int io_tlate<gsl_root_brent<void *,funct<void *> > >::input
    (cinput *co, in_file_format *ins, 
     gsl_root_brent<void *, funct<void *> > *ro);
  template<> int io_tlate<gsl_root_brent<void *,funct<void *> > >::output
    (coutput *co, out_file_format *outs, 
     gsl_root_brent<void *, funct<void *> > *ro);
  template<> const char *io_tlate<gsl_root_brent<void *,
    funct<void *> > >::type();
#endif
  
  typedef io_tlate<gsl_root_brent<void *,funct<void *> > > 
    gsl_root_brent_io_type;
 
#ifndef DOXYGENP
}
#endif

#endif
